#!/usr/bin/python
# Copyright 2014-2016 Nervana Systems Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import division
import numpy           as np
import pycuda.driver   as drv
from neon import logger as neon_logger
from neon.backends.nervanagpu import NervanaGPU
from neon.backends.layer_gpu  import Layer, DataLayer, ConvLayer, PoolLayer, FullLayer, Inception, BatchNorm

# Compare results here:
# https://github.com/soumith/convnet-benchmarks

# Available nets:
# "Alexnet","Overfeat","GoogLeNet1","GoogLeNet2","VGG","VGG_E"
# Note GoogLeNet2 only fits in fp16 currently.  I need to work out delta sharing in inception layers.
nets = ("Alexnet","AlexnetBN","GoogLeNet1BN",)

#Available dtypes: np.float16, np.float32
dtypes = (np.float16, )

# number of full iterations
loops       = 10
# show bechmark details for each layer
layer_bench = 0
# show layer stats after each operation
print_stats = 0
# run network with all zeros to see speed difference
zeros       = 0
# print more stuff
verbose     = 0

ng = NervanaGPU(bench=layer_bench)

neon_logger.display(drv.Context.get_current().get_device().name())

# common convolutional layer settings
conv11    = { "R":11, "S":11, "pad_h":2, "pad_w":2, "str_h":4, "str_w":4 }
conv11p0  = { "R":11, "S":11, "pad_h":0, "pad_w":0, "str_h":4, "str_w":4 }
conv7     = { "R":7,  "S":7,  "pad_h":3, "pad_w":3, "str_h":2, "str_w":2 }
conv5     = { "R":5,  "S":5,  "pad_h":2, "pad_w":2 }
conv5p0   = { "R":5,  "S":5,  "pad_h":0, "pad_w":0 }
conv3     = { "R":3,  "S":3,  "pad_h":1, "pad_w":1 }
conv2     = { "R":2,  "S":2,  "pad_h":0, "pad_w":0, "str_h":2, "str_w":2 }
conv1     = { "R":1,  "S":1,  "pad_h":0, "pad_w":0 }

# traditional pooling
pool2s2p0 = { "R":2, "S":2 }
pool3s2p0 = { "R":3, "S":3, "str_h":2, "str_w":2 }
pool3s2p1 = { "R":3, "S":3, "str_h":2, "str_w":2, "pad_h":1, "pad_w":1 }
pool3s1p1 = { "R":3, "S":3, "str_h":1, "str_w":1, "pad_h":1, "pad_w":1 }
pool7s1p0 = { "R":7, "S":7, "str_h":1, "str_w":1 }

# maxout pooling
pool1j2 = { "op":"max", "J":2 } # maxout in the fc layers
pool2j2 = { "op":"max", "J":2, "R":2, "S":2 }
pool3j2 = { "op":"max", "J":2, "R":3, "S":3 }

def inception1(conf):
    return {
        "layer":Inception, "partitions" : (
            (
                { "layer":ConvLayer, "common":conv1, "relu":True, "K":conf[0][0], },
            ),
            (
                { "layer":ConvLayer, "common":conv1, "relu":True, "K":conf[1][0], },
                { "layer":ConvLayer, "common":conv3, "relu":True, "K":conf[1][1], },
            ),
            (
                { "layer":ConvLayer, "common":conv1, "relu":True, "K":conf[2][0], },
                { "layer":ConvLayer, "common":conv5, "relu":True, "K":conf[2][1], },
            ),
            (
                { "layer":PoolLayer, "common":pool3s1p1, "op":"max" },
                { "layer":ConvLayer, "common":conv1, "relu":True, "K":conf[3][0], },
            ),
        )
    }
def inception1BN(conf):
    return {
        "layer":Inception, "partitions" : (
            (
                { "layer":ConvLayer, "common":conv1, "K":conf[0][0], "bsum":True, },
                { "layer":BatchNorm, "relu":True, "bsum":True },
            ),
            (
                { "layer":ConvLayer, "common":conv1, "K":conf[1][0], "bsum":True, },
                { "layer":BatchNorm, "relu":True, "bsum":True },
                { "layer":ConvLayer, "common":conv3, "K":conf[1][1], "bsum":True, },
                { "layer":BatchNorm, "relu":True, "bsum":True },
            ),
            (
                { "layer":ConvLayer, "common":conv1, "K":conf[2][0], "bsum":True, },
                { "layer":BatchNorm, "relu":True, "bsum":True },
                { "layer":ConvLayer, "common":conv5, "K":conf[2][1], "bsum":True, },
                { "layer":BatchNorm, "relu":True, "bsum":True },
            ),
            (
                { "layer":PoolLayer, "common":pool3s1p1, "op":"max" },
                { "layer":ConvLayer, "common":conv1, "K":conf[3][0], "bsum":True, },
                { "layer":BatchNorm, "relu":True, "bsum":True },
            ),
        )
    }

def inception2(conf):
    layer = { "layer":Inception, "partitions" : [] }
    partitions = layer["partitions"]

    if conf[0][0]:
        partitions.append( (
            { "layer":ConvLayer, "common":conv1, "relu":True, "K":conf[0][0] },
        ) )

    partitions.extend( (
        (
            { "layer":ConvLayer, "common":conv1, "relu":True, "K":conf[1][0] },
            { "layer":ConvLayer, "common":conv3, "relu":True, "K":conf[1][1] },
        ),
        (
            { "layer":ConvLayer, "common":conv1, "relu":True, "K":conf[2][0] },
            { "layer":ConvLayer, "common":conv3, "relu":True, "K":conf[2][1] },
            { "layer":ConvLayer, "common":conv3, "relu":True, "K":conf[2][1] },
        ),
    ) )
    if conf[3][1]:
        partitions.append( (
            { "layer":PoolLayer, "common":pool3s1p1, "op":conf[3][0] }, #
            { "layer":ConvLayer, "common":conv1, "relu":True, "K":conf[3][1] },
        ) )
    else:
        partitions.append( (
            { "layer":PoolLayer, "common":pool3s1p1, "op":conf[3][0] },
        ) )
    return layer

def inception2BN(conf):
    layer = { "layer":Inception, "partitions" : [] }
    partitions = layer["partitions"]

    if conf[0][0]:
        partitions.append( (
            { "layer":ConvLayer, "common":conv1, "K":conf[0][0], "bsum":True },
            { "layer":BatchNorm, "relu":True, "bsum":True },
        ) )

    partitions.extend( (
        (
            { "layer":ConvLayer, "common":conv1, "K":conf[1][0], "bsum":True },
            { "layer":BatchNorm, "relu":True, "bsum":True },
            { "layer":ConvLayer, "common":conv3, "K":conf[1][1], "bsum":True },
            { "layer":BatchNorm, "relu":True, "bsum":True },
        ),
        (
            { "layer":ConvLayer, "common":conv1, "K":conf[2][0], "bsum":True },
            { "layer":BatchNorm, "relu":True, "bsum":True },
            { "layer":ConvLayer, "common":conv3, "K":conf[2][1], "bsum":True },
            { "layer":BatchNorm, "relu":True, "bsum":True },
            { "layer":ConvLayer, "common":conv3, "K":conf[2][1], "bsum":True },
            { "layer":BatchNorm, "relu":True, "bsum":True },
        ),
    ) )
    if conf[3][1]:
        partitions.append( (
            { "layer":PoolLayer, "common":pool3s1p1, "op":conf[3][0] },
            { "layer":ConvLayer, "common":conv1, "K":conf[3][1], "bsum":True },
            { "layer":BatchNorm, "relu":True, "bsum":True },
        ) )
    else:
        partitions.append( (
            { "layer":PoolLayer, "common":pool3s1p1, "op":conf[3][0] },
        ) )
    return layer

networks = {
    "Alexnet" : (
        { "warmup":4 },
        { "layer":DataLayer, "N":128, "C":3, "H":224, "W":224},
        { "layer":ConvLayer, "common":conv11, "relu":True,"K":64  },
        { "layer":PoolLayer, "common":pool3s2p0, "op":"max" },
        { "layer":ConvLayer, "common":conv5, "relu":True, "K":192 },
        { "layer":PoolLayer, "common":pool3s2p0, "op":"max" },
        { "layer":ConvLayer, "common":conv3, "relu":True, "K":384 },
        { "layer":ConvLayer, "common":conv3, "relu":True, "K":256 },
        { "layer":ConvLayer, "common":conv3, "relu":True, "K":256 },
        { "layer":PoolLayer, "common":pool3s2p0, "op":"max" },
        { "layer":FullLayer, "nOut":4096, "relu":True },
        { "layer":FullLayer, "nOut":4096, "relu":True },
        { "layer":FullLayer, "nOut":1000, "relu":True },
    ),
    "AlexnetBN" : (
        { "warmup":4 },
        { "layer":DataLayer, "N":128, "C":3, "H":224, "W":224},
        { "layer":ConvLayer, "common":conv11,"K":64, "bsum":True  },
        { "layer":BatchNorm, "relu":True, "bsum":True },
        { "layer":PoolLayer, "common":pool3s2p0, "op":"max" },
        { "layer":ConvLayer, "common":conv5, "K":192, "bsum":True },
        { "layer":BatchNorm, "relu":True, "bsum":True },
        { "layer":PoolLayer, "common":pool3s2p0, "op":"max" },
        { "layer":ConvLayer, "common":conv3, "K":384, "bsum":True },
        { "layer":BatchNorm, "relu":True, "bsum":True },
        { "layer":ConvLayer, "common":conv3, "K":256, "bsum":True },
        { "layer":BatchNorm, "relu":True, "bsum":True },
        { "layer":ConvLayer, "common":conv3, "K":256, "bsum":True },
        { "layer":BatchNorm, "relu":True, "bsum":True },
        { "layer":PoolLayer, "common":pool3s2p0, "op":"max" },
        { "layer":FullLayer, "nOut":4096 },
        { "layer":BatchNorm, "relu":True },
        { "layer":FullLayer, "nOut":4096 },
        { "layer":BatchNorm, "relu":True },
        { "layer":FullLayer, "relu":True, "nOut":1000 },
    ),
    "Overfeat" : (
        { "warmup":1 },
        { "layer":DataLayer, "N":128, "C":3, "H":231, "W":231},
        { "layer":ConvLayer, "common":conv11p0, "relu":True,"K":96   },
        { "layer":PoolLayer, "common":pool2s2p0, "op":"max" },
        { "layer":ConvLayer, "common":conv5p0, "relu":True, "K":256  },
        { "layer":PoolLayer, "common":pool2s2p0, "op":"max" },
        { "layer":ConvLayer, "common":conv3, "relu":True,   "K":512  },
        { "layer":ConvLayer, "common":conv3, "relu":True,   "K":1024 },
        { "layer":ConvLayer, "common":conv3, "relu":True,   "K":1024 },
        { "layer":PoolLayer, "common":pool2s2p0, "op":"max" },
        { "layer":FullLayer, "nOut":3072, "relu":True },
        { "layer":FullLayer, "nOut":4096, "relu":True },
        { "layer":FullLayer, "nOut":1000, "relu":True },
    ),
    "OverfeatBN" : (
        { "warmup":1 },
        { "layer":DataLayer, "N":128, "C":3, "H":231, "W":231},
        { "layer":ConvLayer, "common":conv11p0,"K":96, "bsum":True   },
        { "layer":BatchNorm, "relu":True, "bsum":True },
        { "layer":PoolLayer, "common":pool2s2p0, "op":"max" },
        { "layer":ConvLayer, "common":conv5p0, "K":256, "bsum":True  },
        { "layer":BatchNorm, "relu":True, "bsum":True },
        { "layer":PoolLayer, "common":pool2s2p0, "op":"max" },
        { "layer":ConvLayer, "common":conv3,   "K":512, "bsum":True  },
        { "layer":BatchNorm, "relu":True, "bsum":True },
        { "layer":ConvLayer, "common":conv3,   "K":1024, "bsum":True },
        { "layer":BatchNorm, "relu":True, "bsum":True },
        { "layer":ConvLayer, "common":conv3,   "K":1024, "bsum":True },
        { "layer":BatchNorm, "relu":True, "bsum":True },
        { "layer":PoolLayer, "common":pool2s2p0, "op":"max" },
        { "layer":FullLayer, "nOut":3072 },
        { "layer":BatchNorm, "relu":True },
        { "layer":FullLayer, "nOut":4096 },
        { "layer":BatchNorm, "relu":True },
        { "layer":FullLayer, "relu":True, "nOut":1000 },
    ),
    # See http://arxiv.org/pdf/1409.1556.pdf for variations
    "VGG" : (
        { "warmup":1 },
        { "layer":DataLayer, "N":64, "C":3, "H":224, "W":224},
        { "layer":ConvLayer, "common":conv3, "relu":True, "K":64 },
        { "layer":PoolLayer, "common":pool2s2p0, "op":"max" },
        { "layer":ConvLayer, "common":conv3, "relu":True, "K":128 },
        { "layer":PoolLayer, "common":pool2s2p0, "op":"max" },
        { "layer":ConvLayer, "common":conv3, "relu":True, "K":256 },
        { "layer":ConvLayer, "common":conv3, "relu":True, "K":256 },
        { "layer":PoolLayer, "common":pool2s2p0, "op":"max" },
        { "layer":ConvLayer, "common":conv3, "relu":True, "K":512 },
        { "layer":ConvLayer, "common":conv3, "relu":True, "K":512 },
        { "layer":PoolLayer, "common":pool2s2p0, "op":"max" },
        { "layer":ConvLayer, "common":conv3, "relu":True, "K":512 },
        { "layer":ConvLayer, "common":conv3, "relu":True, "K":512 },
        { "layer":PoolLayer, "common":pool2s2p0, "op":"max" },
        { "layer":FullLayer, "nOut":4096, "relu":True },
        { "layer":FullLayer, "nOut":4096, "relu":True },
        { "layer":FullLayer, "nOut":1000, "relu":True },
    ),
    # See http://arxiv.org/pdf/1409.1556.pdf for variations
    "VGG_BN" : (
        { "warmup":1 },
        { "layer":DataLayer, "N":64, "C":3, "H":224, "W":224},
        { "layer":ConvLayer, "common":conv3, "K":64, "bsum":True },
        { "layer":BatchNorm, "relu":True, "bsum":True },
        { "layer":PoolLayer, "common":pool2s2p0, "op":"max" },
        { "layer":ConvLayer, "common":conv3, "K":128, "bsum":True },
        { "layer":BatchNorm, "relu":True, "bsum":True },
        { "layer":PoolLayer, "common":pool2s2p0, "op":"max" },
        { "layer":ConvLayer, "common":conv3, "K":256, "bsum":True },
        { "layer":BatchNorm, "relu":True, "bsum":True },
        { "layer":ConvLayer, "common":conv3, "K":256, "bsum":True },
        { "layer":BatchNorm, "relu":True, "bsum":True },
        { "layer":PoolLayer, "common":pool2s2p0, "op":"max" },
        { "layer":ConvLayer, "common":conv3, "K":512, "bsum":True },
        { "layer":BatchNorm, "relu":True, "bsum":True },
        { "layer":ConvLayer, "common":conv3, "K":512, "bsum":True },
        { "layer":BatchNorm, "relu":True, "bsum":True },
        { "layer":PoolLayer, "common":pool2s2p0, "op":"max" },
        { "layer":ConvLayer, "common":conv3, "K":512, "bsum":True },
        { "layer":BatchNorm, "relu":True, "bsum":True },
        { "layer":ConvLayer, "common":conv3, "K":512, "bsum":True },
        { "layer":BatchNorm, "relu":True, "bsum":True },
        { "layer":PoolLayer, "common":pool2s2p0, "op":"max" },
        { "layer":FullLayer, "nOut":4096 },
        { "layer":BatchNorm, "relu":True },
        { "layer":FullLayer, "nOut":4096 },
        { "layer":BatchNorm, "relu":True },
        { "layer":FullLayer, "nOut":1000, "relu":True },
    ),
    # Here is the biggest VGG model (19 layers)
    "VGG_E" : (
        { "warmup":1 },
        { "layer":DataLayer, "N":64, "C":3, "H":224, "W":224},
        { "layer":ConvLayer, "common":conv3, "relu":True, "K":64 },
        { "layer":ConvLayer, "common":conv3, "relu":True, "K":64 },
        { "layer":PoolLayer, "common":pool2s2p0, "op":"max" },
        { "layer":ConvLayer, "common":conv3, "relu":True, "K":128 },
        { "layer":ConvLayer, "common":conv3, "relu":True, "K":128 },
        { "layer":PoolLayer, "common":pool2s2p0, "op":"max" },
        { "layer":ConvLayer, "common":conv3, "relu":True, "K":256 },
        { "layer":ConvLayer, "common":conv3, "relu":True, "K":256 },
        { "layer":ConvLayer, "common":conv3, "relu":True, "K":256 },
        { "layer":ConvLayer, "common":conv3, "relu":True, "K":256 },
        { "layer":PoolLayer, "common":pool2s2p0, "op":"max" },
        { "layer":ConvLayer, "common":conv3, "relu":True, "K":512 },
        { "layer":ConvLayer, "common":conv3, "relu":True, "K":512 },
        { "layer":ConvLayer, "common":conv3, "relu":True, "K":512 },
        { "layer":ConvLayer, "common":conv3, "relu":True, "K":512 },
        { "layer":PoolLayer, "common":pool2s2p0, "op":"max" },
        { "layer":ConvLayer, "common":conv3, "relu":True, "K":512 },
        { "layer":ConvLayer, "common":conv3, "relu":True, "K":512 },
        { "layer":ConvLayer, "common":conv3, "relu":True, "K":512 },
        { "layer":ConvLayer, "common":conv3, "relu":True, "K":512 },
        { "layer":PoolLayer, "common":pool2s2p0, "op":"max" },
        { "layer":FullLayer, "nOut":4096, "relu":True },
        { "layer":FullLayer, "nOut":4096, "relu":True },
        { "layer":FullLayer, "nOut":1000, "relu":True },
    ),
    # http://arxiv.org/abs/1409.4842
    "GoogLeNet1" : (
        { "warmup":1 },
        { "layer":DataLayer, "N":128, "C":3, "H":224, "W":224 },
        { "layer":ConvLayer, "common":conv7, "relu":True, "K":64  },
        { "layer":PoolLayer, "common":pool3s2p1, "op":"max" },
        { "layer":ConvLayer, "common":conv1, "relu":True, "K":64  },
        { "layer":ConvLayer, "common":conv3, "relu":True, "K":192 },
        { "layer":PoolLayer, "common":pool3s2p1, "op":"max" },
        inception1( [(64, ),(96, 128),(16, 32),(32, )] ),
        inception1( [(128,),(128,192),(32, 96),(64, )] ),
        { "layer":PoolLayer, "common":pool3s2p1, "op":"max" },
        inception1( [(192,),(96, 208),(16, 48),(64, )] ),
        inception1( [(160,),(112,224),(24, 64),(64, )] ),
        inception1( [(128,),(128,256),(24, 64),(64, )] ),
        inception1( [(112,),(144,288),(32, 64),(64, )] ),
        inception1( [(256,),(160,320),(32,128),(128,)] ),
        { "layer":PoolLayer, "common":pool3s2p1, "op":"max" },
        inception1( [(256,),(160,320),(32,128),(128,)] ),
        inception1( [(384,),(192,384),(48,128),(128,)] ),
        { "layer":PoolLayer, "common":pool7s1p0, "op":"avg" },
        { "layer":FullLayer, "nOut":1000, "relu":True },
    ),
    # http://arxiv.org/abs/1409.4842
    "GoogLeNet1BN" : (
        { "warmup":1 },
        { "layer":DataLayer, "N":128, "C":3, "H":224, "W":224 },
        { "layer":ConvLayer, "common":conv7, "K":64, "bsum":True  },
        { "layer":BatchNorm, "relu":True, "bsum":True },
        { "layer":PoolLayer, "common":pool3s2p1, "op":"max" },
        { "layer":ConvLayer, "common":conv1, "K":64, "bsum":True  },
        { "layer":BatchNorm, "relu":True, "bsum":True },
        { "layer":ConvLayer, "common":conv3, "K":192, "bsum":True },
        { "layer":BatchNorm, "relu":True, "bsum":True },
        { "layer":PoolLayer, "common":pool3s2p1, "op":"max" },
        inception1BN( [(64, ),(96, 128),(16, 32),(32, )] ),
        inception1BN( [(128,),(128,192),(32, 96),(64, )] ),
        { "layer":PoolLayer, "common":pool3s2p1, "op":"max" },
        inception1BN( [(192,),(96, 208),(16, 48),(64, )] ),
        inception1BN( [(160,),(112,224),(24, 64),(64, )] ),
        inception1BN( [(128,),(128,256),(24, 64),(64, )] ),
        inception1BN( [(112,),(144,288),(32, 64),(64, )] ),
        inception1BN( [(256,),(160,320),(32,128),(128,)] ),
        { "layer":PoolLayer, "common":pool3s2p1, "op":"max" },
        inception1BN( [(256,),(160,320),(32,128),(128,)] ),
        inception1BN( [(384,),(192,384),(48,128),(128,)] ),
        { "layer":PoolLayer, "common":pool7s1p0, "op":"avg" },
        { "layer":FullLayer, "nOut":1000, "relu":True },
    ),
    # adapted from: https://github.com/soumith/kaggle_retinopathy_starter.torch/blob/master/models/googlenet_cudnn.lua
    "GoogLeNet2" : (
        { "warmup":1 },
        { "layer":DataLayer, "N":128, "C":3, "H":224, "W":224 },
        { "layer":ConvLayer, "common":conv7, "relu":True, "K":64  },
        { "layer":PoolLayer, "common":pool3s2p1, "op":"max" },
        { "layer":ConvLayer, "common":conv1, "relu":True, "K":64  },
        { "layer":ConvLayer, "common":conv3, "relu":True, "K":192 },
        { "layer":PoolLayer, "common":pool3s2p1, "op":"max" },
        inception2( [( 64,),( 64, 64),( 64, 96),('avg', 32)] ),
        inception2( [( 64,),( 64, 96),( 64, 96),('avg', 64)] ),
        inception2( [(  0,),(128,160),( 64, 96),('max',  0)] ),
        { "layer":ConvLayer, "common":conv2, "relu":True, "K":576 },
        inception2( [(224,),( 64, 96),( 96,128),('avg',128)] ),
        inception2( [(192,),( 96,128),( 96,128),('avg',128)] ),
        inception2( [(160,),(128,160),(128,160),('avg', 96)] ),
        inception2( [( 96,),(128,192),(160,192),('avg', 96)] ),
        inception2( [(  0,),(128,192),(192,256),('max',  0)] ),
        { "layer":ConvLayer, "common":conv2, "relu":True, "K":1024 },
        inception2( [(352,),(192,320),(160,224),('avg',128)] ),
        inception2( [(352,),(192,320),(192,224),('max',128)] ),
        { "layer":PoolLayer, "common":pool7s1p0, "op":"avg" },
        { "layer":FullLayer, "nOut":1000, "relu":True },
    ),
    # adapted from: https://github.com/soumith/kaggle_retinopathy_starter.torch/blob/master/models/googlenet_cudnn.lua
    "GoogLeNet2BN" : (
        { "warmup":1 },
        { "layer":DataLayer, "N":128, "C":3, "H":224, "W":224 },
        { "layer":ConvLayer, "common":conv7, "K":64, "bsum":True  },
        { "layer":BatchNorm, "relu":True, "bsum":True },
        { "layer":PoolLayer, "common":pool3s2p1, "op":"max" },
        { "layer":ConvLayer, "common":conv1, "K":64, "bsum":True  },
        { "layer":BatchNorm, "relu":True, "bsum":True },
        { "layer":ConvLayer, "common":conv3, "K":192, "bsum":True },
        { "layer":BatchNorm, "relu":True, "bsum":True },
        { "layer":PoolLayer, "common":pool3s2p1, "op":"max" },
        inception2BN( [( 64,),( 64, 64),( 64, 96),('avg', 32)] ),
        inception2BN( [( 64,),( 64, 96),( 64, 96),('avg', 64)] ),
        inception2BN( [(  0,),(128,160),( 64, 96),('max',  0)] ),
        { "layer":ConvLayer, "common":conv2, "K":576, "bsum":True },
        { "layer":BatchNorm, "relu":True, "bsum":True },
        inception2BN( [(224,),( 64, 96),( 96,128),('avg',128)] ),
        inception2BN( [(192,),( 96,128),( 96,128),('avg',128)] ),
        inception2BN( [(160,),(128,160),(128,160),('avg', 96)] ),
        inception2BN( [( 96,),(128,192),(160,192),('avg', 96)] ),
        inception2BN( [(  0,),(128,192),(192,256),('max',  0)] ),
        { "layer":ConvLayer, "common":conv2, "K":1024, "bsum":True },
        { "layer":BatchNorm, "relu":True, "bsum":True },
        inception2BN( [(352,),(192,320),(160,224),('avg',128)] ),
        inception2BN( [(352,),(192,320),(192,224),('max',128)] ),
        { "layer":PoolLayer, "common":pool7s1p0, "op":"avg" },
        { "layer":FullLayer, "nOut":1000, "relu":True },
    ),
}

for net in nets:

    for dtype in dtypes:

        warmup  = networks[net][0]["warmup"]
        network = networks[net][1:]
        name    = "%s (dtype=%s, N=%d)" % (net, np.dtype(dtype).name, network[0]["N"])

        # only first run needs a big warmup
        networks[net][0]["warmup"] = 1

        neon_logger.display("------------------------------------------------")
        neon_logger.display("Benchmarking: " + name)
        neon_logger.display("------------------------------------------------")

        layers = []
        prev_layer = None
        max_deltas  = 0
        max_weights = 0
        max_delta_layer  = None
        max_weight_layer = None
        shared_weights   = None
        shared_deltas    = []
        inception        = False

        for conf in network:

            layer = Layer.create(ng, conf, prev_layer, dtype)

            if type(layer) is Inception:
                inception = True

            # find the size of the largest buffers so they can be shared
            if layer.sizeF > max_weights:
                max_weights = layer.sizeF
                max_weight_layer = layer

            if layer.sizeI > max_deltas and type(prev_layer) is not DataLayer:
                max_deltas = layer.sizeI
                max_delta_layer = layer

            prev_layer = layer
            layers.append(layer)

        # Init shared buffers (assumes consistent dtype for now)
        shared_deltas.append(ng.empty(max_delta_layer.dimI, dtype=max_delta_layer.dtype))
        shared_deltas.append(ng.empty(max_delta_layer.dimI, dtype=max_delta_layer.dtype))
        if inception:
            shared_deltas.append(ng.empty(max_delta_layer.dimI, dtype=max_delta_layer.dtype))
            shared_deltas.append(ng.empty(max_delta_layer.dimI, dtype=max_delta_layer.dtype))

        shared_updates = ng.empty(max_weight_layer.dimF, dtype=np.float32)

        for i, layer in enumerate(layers):
            if verbose:
                neon_logger.display(layer)

            # Intitalize buffers.  Alernate shared delta buffer.
            # One layer can't have the same buffer for both error in and error out.
            layer.init_activations()
            layer.init_weights(shared=shared_updates, zeros=zeros)
            if i > 1:
                layer.init_deltas(shared=shared_deltas)

        if verbose:
            remain, total = drv.mem_get_info()
            neon_logger.display("%.3fGB of %.3fGB Allocated (%.3fGB Remaining)" %
                  ((total - remain) / 1024.**3, total / 1024.**3, remain / 1024.**3))

        if zeros:
            layers[0].init_data()
        else:
            # give the first layer some data
            layers[0].init_data(np.random.uniform(0.0, 1.0, layers[0].dimO))

            # Scale the initial weights so activations are bound around 1.0
            # We do this by running it through the forward pass and collecting mean stats
            # ng.bench = False
            # propagation = None
            # for layer in layers:
            #     propagation = layer.fprop(propagation, scale_weights=.5)
            # ng.bench = layer_bench

        start = drv.Event()
        end   = drv.Event()

        fprop_time  = 0
        bprop_time  = 0
        fprop_flops = 0
        bprop_flops = 0

        # We throw away the first two runs as it includes pycuda kernel loading times and clock warmup.
        # So add 1 to our loop count.
        for loop in range(loops + warmup):

            loop = loop - warmup + 1
            if loop < 0: loop = 0

            start.record()
            flops = 0

            #fprop
            propagation = None
            for layer in layers:

                propagation = layer.fprop(propagation)

                flops += layer.flops
                if print_stats:
                    layer.fprop_stats()

            end.record()
            end.synchronize()
            msecs = end.time_since(start)
            neon_logger.display("fprop(%2d): %8.3f msecs %8.3f gflops" %
                  (loop, msecs, flops / (msecs * 1000000.0)))
            if loop > 0:
                fprop_time  += msecs
                fprop_flops += flops

            start.record()
            flops = 0

            #bprop
            for layer in layers[:0:-1]:

                propagation = layer.bprop(propagation)

                flops += layer.flops * 2
                if print_stats:
                    layer.bprop_stats()

            end.record()
            end.synchronize()
            msecs = end.time_since(start)
            neon_logger.display("bprop(%2d): %8.3f msecs %8.3f gflops" %
                  (loop, msecs, flops / (msecs * 1000000.0)))
            if loop > 0:
                bprop_time  += msecs
                bprop_flops += flops

        if loops > 0:

            neon_logger.display("---------------------------------------------")
            neon_logger.display(name + " Results:")
            neon_logger.display("---------------------------------------------")
            neon_logger.display("Avg(%d) fprop: %8.3f msecs %.3f gflops" %
                  (loops, fprop_time / loops, fprop_flops / (fprop_time * 1000000.0)))

            neon_logger.display("Avg(%d) bprop: %8.3f msecs %.3f gflops" %
                  (loops, bprop_time / loops, bprop_flops / (bprop_time * 1000000.0)))

            fprop_time  += bprop_time
            fprop_flops += bprop_flops

            neon_logger.display("Avg(%d) total: %8.3f msecs %.3f gflops\n\n" %
                  (loops, fprop_time / loops, fprop_flops / (fprop_time * 1000000.0)))



# ----------------------------------------------------------------------------
# Copyright 2017 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
"""
MKL based backend interface and tensor data structure.
"""
from __future__ import division
from builtins import zip
import logging

from neon.backends import layer_mkl
from neon.backends.util.check_mkl import get_mkl_lib
from neon.backends.nervanacpu import CPUTensor, NervanaCPU, CustomNumpy
import ctypes
from cffi import FFI
from ctypes import c_longlong, c_float, c_double, c_int
import numpy as np
from neon.backends import math_cpu
from neon.backends.backend import OpTreeNode
import os
import sys

_none_slice = slice(None, None, None)

logger = logging.getLogger(__name__)


# TODO: enable this flag to find numerical problems
# np.seterr(all='raise')


class MKLTensor(CPUTensor):
    """
    MKLTensor, special for MKL Backend
    """
    _tensor = None

    def __init__(self,
                 backend,
                 shape=None,
                 dtype=np.float32,
                 ary=None,
                 name=None,
                 persist_values=True,
                 base=None):

        super(MKLTensor, self).__init__(backend, shape, dtype, ary, name,
                                        persist_values, base)

        # add 4 address for cpu layout and buffer, mkl layout and buffer
        self.primitive = np.zeros((4), dtype=np.uint64)
        if ary is not None:
            self.primitive[1] = self._tensor.ctypes.data
        self.shape5D = None

    def get_prim(self):
        return c_longlong(self.primitive.ctypes.data)

    def clean_mkl(self):
        self.primitive[2] = 0
        self.primitive[3] = 0

    def set_mkl(self, tensor):
        self.primitive[0] = tensor.primitive[0]
        self.primitive[2] = tensor.primitive[2]
        self.primitive[3] = tensor.primitive[3]
        self.shape5D = tensor.shape5D

    def __str__(self):
        """
        Returns a string representation of this Tensor.

        Returns:
            str: the representation.
        """
        if self._tensor.base is not None:
            base_id = id(self._tensor.base)
        else:
            base_id = id(self._tensor)
        return ("MKLTensor(base 0x%x) name:%s shape:%s dtype:%s strides:%s"
                " is_c_contiguous:%s" % (base_id, self.name, self.shape,
                                         self.dtype, self._tensor.strides,
                                         self._tensor.flags.c_contiguous))

    def get(self):
        """
        Return the array.
        """
        self.backend.convert(self)
        return self._tensor.copy()

    def reshape(self, *shape):
        """
        Return a reshaped view.
        """
        newTensor = super(MKLTensor, self).reshape(*shape)
        newTensor.set_mkl(self)
        return newTensor


def _assign_right_to_left(left, right):
    math_cpu.blas_copy(left, right)


# how to overlaod numpy_call_dict?
numpy_call_dict_mkl = {
    # assign
    "assign": _assign_right_to_left,
    # zero_operand ops
    # unary ops
    "neg": lambda left: math_cpu.neg(left),
    "abs": lambda left: np.abs(left),
    "sgn": lambda left: np.sign(left),
    "sqrt": lambda left: math_cpu.sqrt(left),
    "sqr": lambda left: math_cpu.square(left),
    "exp": lambda left: math_cpu.exp(left),
    "log": lambda left: math_cpu.log(left),
    "safelog": lambda left: math_cpu.safelog(left),
    "exp2": lambda left: np.exp2(left),
    "log2": lambda left: np.log2(left),
    "sig": lambda left: 1. / (1. + np.exp(-left)),
    "sig2": lambda left: 1. / (1. + np.exp2(-left)),
    "tanh": lambda left: np.tanh(left),
    "tanh2": lambda left: (np.exp2(2. * left) - 1.) / (np.exp2(2. * left) + 1.),
    "transpose": lambda left: np.transpose(left),
    "rint": lambda left: np.rint(left),
    # binary ops
    "add": lambda left, right: math_cpu.add(left, right),
    "sub": lambda left, right: math_cpu.sub(left, right),
    "mul": lambda left, right: math_cpu.mul(left, right),
    "div": lambda left, right: math_cpu.div(left, right),
    "eq": lambda left, right: left == right,
    "ne": lambda left, right: left != right,
    "lt": lambda left, right: left < right,
    "le": lambda left, right: left <= right,
    "gt": lambda left, right: left > right,
    "ge": lambda left, right: left >= right,
    "pow": lambda left, right: np.power(left, right),
    "minimum": lambda left, right: np.minimum(left, right),
    "maximum": lambda left, right: np.maximum(left, right),
    "dot": lambda left, right: np.dot(left, right),
    # reduction ops
    "sum": lambda op_dict, left: math_cpu.sum(left, axis=op_dict['axis'], keepdims=True),
    "max": lambda op_dict, left: np.max(left, axis=op_dict['axis'], keepdims=True),
    "min": lambda op_dict, left: np.min(left, axis=op_dict['axis'], keepdims=True),
    "argmax": lambda op_dict, left: CustomNumpy.argmax(left, axis=op_dict['axis'], keepdims=True),
    "argmin": lambda op_dict, left: CustomNumpy.argmin(left, axis=op_dict['axis'], keepdims=True),
}


class NervanaMKL(NervanaCPU):
    """
    MKL Backend
    """
    backend_name = 'mkl'

    def __init__(self,
                 rng_seed=None,
                 default_dtype=np.float32,
                 hist_bins=64,
                 hist_offset=-48,
                 compat_mode=None,
                 # Ignored
                 num_devices=None,
                 stochastic_round=None,
                 device_id=None,
                 deterministic=None
                 ):
        super(NervanaMKL, self).__init__(rng_seed, default_dtype,
                                         hist_bins, hist_offset, compat_mode=compat_mode)
        self.tensor_cls = MKLTensor
        logger.info("Initialized NervanaMKL")
        assert get_mkl_lib(), "MKL is not installed correctly"

        path = os.path.dirname(os.path.realpath(__file__))
        header_path = os.path.join(os.path.dirname(__file__), 'mklEngine', 'src',
                                   'math_cpu.header')

        if sys.platform == 'win32':
            mkl_ml_path = os.path.join(path, os.pardir, 'backends', 'mklEngine', 'mklml.dll')
            ctypes.windll.LoadLibrary(mkl_ml_path)
            mkl_engine_path = os.path.join(path, os.pardir, 'backends', 'mklEngine',
                                           'mklEngine.dll')
            self.mklEngine = ctypes.windll.LoadLibrary(mkl_engine_path)
            math_engine_path = os.path.join(os.path.dirname(__file__), 'mklEngine', 'cmath.dll')

        elif sys.platform == 'darwin':
            mkl_ml_path = os.path.join(path, os.pardir, 'backends', 'mklEngine',
                                       'libmklml.dylib')
            ctypes.cdll.LoadLibrary(mkl_ml_path)
            iomp5_path = os.path.join(path, os.pardir, 'backends', 'mklEngine',
                                      'libiomp5.dylib')
            ctypes.cdll.LoadLibrary(iomp5_path)
            mkl_engine_path = os.path.join(path, os.pardir, 'backends', 'mklEngine',
                                           'mklEngine.dylib')
            self.mklEngine = ctypes.cdll.LoadLibrary(mkl_engine_path)
            math_engine_path = os.path.join(os.path.dirname(__file__), 'mklEngine', 'cmath.dylib')

        else:
            mkl_engine_path = os.path.join(path, os.pardir, 'backends', 'mklEngine',
                                           'mklEngine.so')
            self.mklEngine = ctypes.cdll.LoadLibrary(mkl_engine_path)
            math_engine_path = os.path.join(os.path.dirname(__file__), 'mklEngine', 'cmath.so')

        ffi = FFI()
        with open(header_path) as header:
            ffi.cdef(header.read())
        self.mathlib = ffi.dlopen(math_engine_path)

    def is_mkl(self):
        return True

    def convert(self, a):
        if a.primitive[2] == 0:
            return a
        C, D, H, W, N = a.shape5D
        self.mklEngine.ConvertBack(a.get_prim(), N, C, H, W)
        return a

    def convert_mkl(self, a):
        self.mklEngine.ConvertToMKL(a.get_prim())
        return a

    def convert_data(self, tensor, layer_mkl):
        if not layer_mkl and tensor is not None:
            if type(tensor) == MKLTensor:
                self.convert(tensor)
                tensor.clean_mkl()
            elif type(tensor) is list or type(tensor) is tuple:
                for i in tensor:
                    self.convert_data(i, layer_mkl)
            elif type(tensor) is OpTreeNode or type(tensor) is np.ndarray:
                return
            else:
                assert False, 'unsupported input for convert ' + str(type(tensor))

    def clean_data(self, tensor, layer_mkl):
        if layer_mkl and tensor is not None and type(tensor) == MKLTensor:
            tensor.clean_mkl()

    def get_numpy(self, a):
        # transfer mkl tensor into a new numpy
        numpy_a = a._tensor.copy()
        b = a.primitive[1]
        a.primitive[1] = numpy_a.ctypes.data
        self.convert(a)
        a.primitive[1] = b
        return numpy_a

    def execute(self, optree):
        return super(NervanaMKL, self).execute(optree, numpy_call_dict=numpy_call_dict_mkl)

    def copy_transpose(self, a, out, axes=None, repeat=1):
        """
        use MKL transposition to speed up
        """
        if axes is None and a._tensor.ctypes.data != out._tensor.ctypes.data and len(a.shape) == 2:
            inp = c_longlong(a._tensor.ctypes.data)
            outp = c_longlong(out._tensor.ctypes.data)
            m, n = a.shape
            self.mklEngine.MatTrans(inp, outp, c_longlong(m), c_longlong(n))
        else:
            out._tensor[:] = np.transpose(a._tensor, axes).copy()

    def conv_layer(self, dtype,
                   N, C, K,
                   D=1, H=1, W=1,
                   T=1, R=1, S=1,
                   pad_d=0, pad_h=0, pad_w=0,
                   str_d=1, str_h=1, str_w=1,
                   dil_d=1, dil_h=1, dil_w=1):
        """
        Create a new ConvLayer parameter object.
        This then is passed as an argument to all the convolution operations.

        N: Number of images in mini-batch
        C: Number of input feature maps
        K: Number of output feature maps

        D: Depth  of input image
        H: Height of input image
        W: Width  of input image

        T: Depth  of filter kernel
        R: Height of filter kernel
        S: Width  of filter kernel

        padding: amount of zero-padding around the given edge
        strides: factor to step the filters by in a given direction
        dilation: dilation factor for each dimension

        dtype: need to know dtype to setup proper kernels and params.

        bsum: calculate the sum along the batchnorm axis for fprop or bprop
              outputs an fp32 tensor of size Kx1

        """
        return layer_mkl.ConvLayerMKL(
            self, dtype, N, C, K, D, H, W, T, R, S,
            pad_d, pad_h, pad_w, str_d, str_h, str_w,
            dil_d, dil_h, dil_w)

    def deconv_layer(self, dtype,
                     N, C, K,
                     M, P, Q,
                     T=1, R=1, S=1,
                     pad_d=0, pad_h=0, pad_w=0,
                     str_d=1, str_h=1, str_w=1,
                     dil_d=1, dil_h=1, dil_w=1):
        """
        Create a new DeconvLayer parameter object.
        This then is passed as an argument to all the convolution operations.

        N: Number of images in mini-batch
        C: Number of output feature maps
        K: Number of input feature maps

        M: Depth  of input
        P: Height of input
        Q: Width of input

        D: Depth  of output image
        H: Height of output image
        W: Width  of output image

        T: Depth  of filter kernel
        R: Height of filter kernel
        S: Width  of filter kernel

        padding: amount of zero-padding around the given edge
        strides: factor to step the filters by in a given direction
        dilation: dilation factor for each dimension

        dtype: need to know dtype to setup proper kernels and params.
        """
        return layer_mkl.DeconvLayerMKL(self, dtype, N, C, K, M, P, Q, T, R, S,
                                        pad_d, pad_h, pad_w, str_d, str_h, str_w,
                                        dil_d, dil_h, dil_w)

    def pool_layer(self, dtype,
                   op, N, C,
                   D=1, H=1, W=1,
                   J=1, T=1, R=1, S=1,
                   pad_c=0, pad_d=0, pad_h=0, pad_w=0,
                   str_c=None, str_d=None, str_h=None, str_w=None):
        """
        Create a new PoolLayer parameter object.
        This then is passed as an argument to all pooling kernels.

        op: "max", "avg", "l2" pooling (currently bprop only supports max, but not avg and l2)
        N: Number of images in mini-batch

        C: Number of input feature maps
        D: Depth  of input image
        H: Height of input image
        W: Width  of input image

        J: Size of feature map pooling window (maxout n_pieces)
        T: Depth  of pooling window
        R: Height of pooling window
        S: Width  of pooling window

        padding: amount of zero-padding around the given image or feature map edge
        strides: factor to step the window by in a given direction (overlap allowed)

        Leave spatial dimensions at 1 to allow feature map pooling in the fc layers.
        """
        # default to non-overlapping
        if str_c is None:
            str_c = J
        if str_d is None:
            str_d = T
        if str_h is None:
            str_h = R
        if str_w is None:
            str_w = S

        return layer_mkl.PoolLayerMKL(self, dtype, op, N, C, D, H, W, J, T, R, S,
                                      pad_c, pad_d, pad_h, pad_w, str_c, str_d, str_h, str_w)

    def fprop_pool(self, layer, I, O, argmax=None, beta=0.0):
        """
        Forward propagate pooling layer.

        Arguments:
            layer (PoolLayer): The pool layer object, different backends have
                               different pool layers.
            I (Tensor): Input tensor.
            O (Tensor): output tensor.
            argmax (Tensor): tensor to store location of the maximum
        """
        assert layer.sizeI == I.size
        assert layer.sizeO == O.size
        assert layer.op == "max" or layer.op == 'avg'
        if layer.op == "max":
            assert layer.sizeO == argmax.size

        J, T, R, S = layer.JTRS
        C, D, H, W, N = layer.dimI
        K, M, P, Q, N = layer.dimO
        pad_c, pad_d, pad_h, pad_w = layer.padding
        str_c, str_d, str_h, str_w = layer.strides

        # unsupported fall back to cpu
        if J > 1 or T > 1 or D > 1:
            super(NervanaMKL, self).fprop_pool(layer, I, O, argmax, beta)
            return

        if layer.op == "max":
            bMax = 1
        elif layer.op == 'avg':
            bMax = 0
        if self.check_caffe_compat():
            bCeil = 1
        else:
            bCeil = 0

        primitives = c_longlong(layer.dnnPrimitives.ctypes.data)
        self.mklEngine.MaxPooling_fprop(
            I.get_prim(), O.get_prim(), primitives, layer.initOk_f, bMax,
            N, C, H, W, R, S, str_h, str_w, pad_h, pad_w, K, P, Q, bCeil)
        layer.initOk_f = 1
        O.shape5D = layer.dimO

    def bprop_pool(self, layer, I, O, argmax=None, alpha=1.0, beta=0.0):
        """
        Backward propagate pooling layer.

        Arguments:
            layer (PoolLayer): The pool layer object. Different backends have
                               different pool layers.
            I (Tensor): Input (error) tensor.
            O (Tensor): Output (delta) tensor.
            argmax (Tensor): tensor to store location of the maximum
            alpha (float): linear scaling (does not work for l2 pooling)
            beta (float): accumulation value into grad_I
        """
        assert layer.sizeI == O.size
        assert layer.sizeO == I.size
        if layer.op == "max":
            assert layer.sizeO == argmax.size

        J, T, R, S = layer.JTRS
        C, D, H, W, N = layer.dimI
        K, M, P, Q, N = layer.dimO
        pad_c, pad_d, pad_h, pad_w = layer.padding
        str_c, str_d, str_h, str_w = layer.strides

        # unsupported fall back to cpu
        if J > 1 or T > 1 or D > 1:
            super(NervanaMKL, self).bprop_pool(layer, I, O, argmax, alpha, beta)
            return

        primitives = c_longlong(layer.dnnPrimitives.ctypes.data)
        self.mklEngine.MaxPooling_bprop(I.get_prim(), O.get_prim(),
                                        primitives, layer.initOk_b, c_float(beta))
        layer.initOk_b = 1
        O.shape5D = layer.dimI

    def compound_dot(self, A, B, C, alpha=1.0, beta=0.0, relu=False, bsum=None):
        """
        Doing following operations (* is dot product)
        C = alpha * A * B   + beta * C
        C = alpha * A.T * B + beta * C
        C = alpha * A * B.T + beta * C.

        relu: if true applied before output (and prior to beta addition)

        The operation will be short-circuited to: out <- alpha * left * right
        if beta has value 0 (the default).

        Arguments:
            A, B (CPUTensor): input operands
            C (MCPUTensor): output
            alpha (float): scale A*B term
            beta (float): scale C term before sum
            relu (bool): whether to apply ReLu before output
        """

        # checking type and shape
        assert A.dtype == B.dtype == C.dtype
        assert A.shape[0] == C.shape[0]
        assert B.shape[1] == C.shape[1]
        assert A.shape[1] == B.shape[0]

        # cleaner implementation, shall be equivalent to the one below
        # if relu:
        #     C[:] = self.log(1. + self.exp(alpha * self.dot(A, B))) + beta * C
        # else:
        #     C[:] = alpha * self.dot(A, B) + beta * C

        if not relu:
            if C._tensor.flags['C_CONTIGUOUS'] is not True:
                tmp = np.empty(C.shape, dtype=C.dtype)
                if beta != 0:
                    tmp[:] = C._tensor
                math_cpu.blas_dot(A._tensor, B._tensor, tmp, alpha, beta)
                C._tensor[:] = tmp
            else:
                math_cpu.blas_dot(A._tensor, B._tensor, C._tensor, alpha, beta)
        else:
            # mfma: change np.multiply to mul
            if beta != 1:
                np.multiply(C._tensor, beta, C._tensor)
            tmp = np.empty(C.shape, dtype=C.dtype)
            np.dot(A._tensor, B._tensor, tmp)
            # mfma: change np.multiply to mul
            if alpha != 1:
                np.multiply(tmp, alpha, tmp)
            if relu:
                self.Relu(tmp, tmp)
            np.add(C._tensor, tmp, C._tensor)
        if bsum is not None:
            bsum[:] = self.sum(C, 1)

        return C

    def relu_layer(self):
        return layer_mkl.ReluLayerMKL()

    def fprop_relu(self, layer, x, slope):
        if layer is None:
            layer = layer_mkl.ReluLayerMKL()
        if not hasattr(x, 'shape5D'):
            return self.maximum(x, 0) + slope * self.minimum(0, x)
        if slope != 0:
            self.convert(x)
            x.clean_mkl()
            return self.maximum(x, 0) + slope * self.minimum(0, x)

        if x.shape5D is not None:
            C, D, H, W, N = x.shape5D
        else:
            C, N = x._tensor.shape
            D, H, W = 1, 1, 1
            x.shape5D = C, D, H, W, N
        layer.shape5D = C, D, H, W, N

        primitives = c_longlong(layer.dnnPrimitives.ctypes.data)
        if x.primitive[3] == 0:
            layer.inputMKL = False
        self.mklEngine.Relu_f(x.get_prim(), primitives, layer.initOk_f, N, C, H, W)
        layer.initOk_f = 1
        return x

    def bprop_relu(self, layer, x, error, deltas, slope):
        if layer is None:
            layer = layer_mkl.ReluLayerMKL()
        if slope != 0 or error is None:
            if error is not None:
                self.convert(error)
                error.clean_mkl()
            return self.greater(x, 0) + slope * self.less(x, 0)

        # to be moved to C code
        if not layer.inputMKL:
            self.convert(error)
            error.clean_mkl()
        primitives = c_longlong(layer.dnnPrimitives.ctypes.data)
        self.mklEngine.Relu_b(x.get_prim(), error.get_prim(), primitives, layer.initOk_b)
        layer.initOk_b = 1
        deltas.set_mkl(error)
        deltas.shape5D = layer.shape5D
        if deltas.primitive[3] == 0:
            deltas[:] = error

    def fprop_transform(self, nglayer, transform, inputs, outputs, relu=False):
        if relu:
            transform(inputs, nglayer)
        else:
            return super(NervanaMKL, self).fprop_transform(
                nglayer, transform, inputs, outputs, relu)

    def bprop_transform(self, nglayer, transform, outputs, error, deltas, relu):
        if relu:
            transform.bprop(outputs, nglayer, error, deltas)
        else:
            super(NervanaMKL, self).bprop_transform(
                nglayer, transform, outputs, error, deltas, relu)

    def compound_fprop_bn(self, x, xsum, xvar, gmean, gvar, gamma, beta, y,
                          eps, rho, compute_batch_sum,
                          accumbeta=0.0, relu=False, binary=False,
                          inference=False, outputs=None, layer=None):
        if layer is None or outputs is None or not isinstance(layer.in_shape, tuple):
            super(NervanaMKL, self).compound_fprop_bn(x, xsum, xvar, gmean, gvar,
                                                      gamma, beta, y, eps, rho,
                                                      compute_batch_sum,
                                                      accumbeta, relu, binary,
                                                      inference, outputs, layer
                                                      )
            return

        primitives = c_longlong(layer.dnnPrimitives.ctypes.data)
        if len(layer.in_shape) == 3:
            C = layer.in_shape[0]
            H = layer.in_shape[1]
            W = layer.in_shape[2]
            D = 1
        elif len(layer.in_shape) == 2:
            C = layer.in_shape[0]
            H = layer.in_shape[1]
            W = layer.in_shape[1]
            D = 1
        elif len(layer.in_shape) == 4:
            C = layer.in_shape[0]
            H = layer.in_shape[1]
            W = layer.in_shape[2]
            D = layer.in_shape[3]

        N = int(x.shape[-1]) // H // W // D  # this is/corresponds to the batch size
        gmean = c_longlong(gmean._tensor.ctypes.data)
        gvar = c_longlong(gvar._tensor.ctypes.data)
        self.mklEngine.BatchNormFprop(x.get_prim(), outputs.get_prim(),
                                      gamma.get_prim(), beta.get_prim(), gmean, gvar,
                                      c_float(rho), N, C, H, W * D, c_double(eps),
                                      primitives, layer.init_f, c_int(inference))
        layer.init_f = 1
        layer.shape5D = outputs.shape5D = C, D, H, W, N

    def batchnorm_layer(self, in_shape):
        return layer_mkl.BatchNormLayerMKL(in_shape)

    def compound_bprop_bn(self, deltas, grad_gamma, grad_beta,
                          error, inputs, xsum, xvar, gamma,
                          eps, binary=False, layer=None):
        if not layer or not isinstance(layer.in_shape, tuple):
            super(NervanaMKL, self).compound_bprop_bn(deltas, grad_gamma,
                                                      grad_beta, error,
                                                      inputs, xsum, xvar,
                                                      gamma, eps, binary,
                                                      layer)
            return

        primitives = c_longlong(layer.dnnPrimitives.ctypes.data)

        self.mklEngine.BatchNormBackp(inputs.get_prim(), error.get_prim(),
                                      deltas.get_prim(), grad_gamma.get_prim(),
                                      grad_beta.get_prim(), layer.in_shape[0],
                                      primitives, layer.init_b)
        layer.init_b = 1
        deltas.shape5D = layer.shape5D

    def fprop_skipnode(self, x, y, beta):
        y.set_mkl(x)
        # skipnode will do nothing for mkltensor but copy for CPU tensor
        if not x.primitive[2]:
            super(NervanaMKL, self).fprop_skipnode(x, y, beta)

    def bprop_skipnode(self, error, delta, alpha, beta):
        delta.set_mkl(error)
        if not error.primitive[2]:
            super(NervanaMKL, self).bprop_skipnode(error, delta, alpha, beta)

    def mergesum_layer(self, layer_num):
        return layer_mkl.MergeSumLayerMKL(layer_num)

    def sum_tensor(self, sum, layer_num, tensors, output):
        inp = c_longlong(tensors.ctypes.data)
        size = c_longlong(np.prod(output.shape))
        prim = c_longlong(sum.ctypes.data)
        self.mklEngine.MklSumTensor(layer_num, inp, size, output.get_prim(), prim)

    def fprop_mergesum(self, ngLayer, inputs, inference, layers, outputs, out_shape):
        ngLayer.shape5D = inputs.shape5D
        for i, l in enumerate(layers):
            l.fprop(inputs, inference)
            alloc_layers = [ll for ll in l.layers if ll.owns_output]
            ngLayer.tensors[(i * 4):(i * 4 + 4)] = alloc_layers[-1].outputs.primitive[0:4]

        C, H, W = out_shape
        outputs.shape5D = (C, 1, H, W, outputs.shape[-1])
        self.sum_tensor(ngLayer.sum_prim_f, ngLayer.layer_num, ngLayer.tensors, outputs)

    def bprop_mergesum(self, ngLayer, alpha, beta, layers, error, deltas):
        for i, l in enumerate(reversed(layers)):
            e = l.bprop(error)
            ngLayer.tensors[(i * 4):(i * 4 + 4)] = e.primitive[0:4]

        self.sum_tensor(ngLayer.sum_prim_b, ngLayer.layer_num, ngLayer.tensors, deltas)
        deltas.shape5D = ngLayer.shape5D

    def mergebroadcast_layer(self, layer_num):
        return layer_mkl.MergeBroadcastLayerMKL(layer_num)

    def fprop_mergebroadcast(self, ngLayer, inputs, inference, outputs, layers, out_shape):
        for l in layers:
            l.fprop(inputs, inference)

        if ngLayer.initOK_f == 0:
            C, H, W = layers[0].out_shape
            N = outputs.shape[-1]
            C = out_shape[0]
            ngLayer.in_shape5D = inputs.shape5D
            ngLayer.out_shape5D = N, C, H, W
            for i, layer in enumerate(layers):
                ngLayer.channels[i] = layer.out_shape[0]

        N, C, H, W = ngLayer.out_shape5D
        outputs.shape5D = C, 1, H, W, N

        for i, layer in enumerate(layers):
            alloc_layers = [l for l in layer.layers if l.owns_output]
            ngLayer.tensors_temp[(i*4):(i*4 + 4)] = alloc_layers[-1].outputs.primitive[0:4]

        channel = c_longlong(ngLayer.channels.ctypes.data)
        inp = c_longlong(ngLayer.tensors_temp.ctypes.data)
        out = outputs.get_prim()
        prim = c_longlong(ngLayer.primitive.ctypes.data)
        self.mklEngine.Concat_f(inp, ngLayer.layer_num,
                                out, prim, channel, ngLayer.initOK_f, N, C, H, W)
        ngLayer.initOK_f = 1

    def bprop_mergebroadcast(self, ngLayer, layers, error_views,
                             error, deltas, out_shape, alpha, beta, alphas, betas):
        C, D, H, W, N = ngLayer.in_shape5D
        i = 0
        for l, e in zip(layers, error_views):
            ngLayer.tensors_temp[(i * 4):(i * 4 + 4)] = e.primitive[0:4]
            i += 1

        channel = c_longlong(ngLayer.channels.ctypes.data)
        tensors = c_longlong(ngLayer.tensors_temp.ctypes.data)
        prim = c_longlong(ngLayer.primitive.ctypes.data)
        self.mklEngine.Concat_b(tensors, ngLayer.layer_num, error.get_prim(), prim, channel,
                                ngLayer.initOK_b, N, H, W)

        ngLayer.initOK_b = 1

        i = 0
        for l, e in list(zip(layers, error_views)):
            e.primitive[0:4] = ngLayer.tensors_temp[(i * 4):(i * 4 + 4)]
            e.shape5D = l.layers[-1].outputs.shape5D
            err = l.bprop(e)
            ngLayer.tensors_temp[(i * 4):(i * 4 + 4)] = err.primitive[0:4]
            i += 1

        if deltas is None:
            return

        size = c_longlong(np.prod(ngLayer.in_shape5D))
        prim = c_longlong(ngLayer.sum_prim.ctypes.data)
        tensors = c_longlong(ngLayer.tensors_temp.ctypes.data)
        self.mklEngine.MklSumTensor(ngLayer.layer_num, tensors, size, deltas.get_prim(), prim)

        deltas.shape5D = ngLayer.in_shape5D

    def compound_rnn_unroll_fprop(self, W_recur, h_prev_s, h_ff_s, h_s, bias,
                                  nout, num_steps, num_used_steps, activation,
                                  reverse=False):
        """
        Time step unrolling portion of recurrent layer fprop.

        Arguments:
            W_recur (Tensor): Recurrent weight matrix.
            h_prev_s (Array): Array of per time step hidden state tensors. Each
                element in the array is a single time step view into one tensor
                containing all of the time steps in sequence.
            h_ff_s (Array): Array of per time step hidden state tensors. Each
                element in the array is a single time step view into one tensor
                containing all of the time steps in sequence.
            h_s (Array): Array of per time step hidden state tensors. Each
                element in the array is a single time step view into one tensor
                containing all of the time steps in sequence.
            bias (Tensor): Bias tensor to add at each time step.
            nout (integer): Number of output units for the layer.
            num_steps (integer): Total number of time steps in the buffer.
            num_used_steps (integer): Number of time steps being used for real
                data.
            activation (Transform): Activation function for the layer.
            reverse (boolean): When true, unrolling will iterate over time steps
                in reverse (for BiRNN).
        """
        if num_used_steps is not None and num_used_steps < num_steps:
            h_s = h_s[:num_used_steps]
            h_prev_s = h_prev_s[:num_used_steps]
            h_ff_s = h_ff_s[:num_used_steps]

        if reverse:
            steps = reversed(list(zip(h_s, h_prev_s, h_ff_s)))
        else:
            steps = zip(h_s, h_prev_s, h_ff_s)

        for (h, h_prev, h_ff) in steps:
            if h_ff is h:
                self.compound_dot(W_recur, h_prev, h, beta=1.0)
                h[:] = activation(h + bias)
            else:
                self.compound_dot(W_recur, h_prev, h)
                if not math_cpu.add_and_act(h._tensor, h_ff._tensor, bias._tensor, activation):
                    h[:] = activation(h + h_ff + bias)

    def compound_rnn_unroll_bprop(self, W_recur, delta_prev_s, delta_s, h_s,
                                  nout, num_steps, num_used_steps, activation,
                                  reverse=True):
        """
        Time step unrolling portion of recurrent layer bprop.

        Arguments:
            W_recur (Tensor): Recurrent weight matrix.
            delta_prev_s (Array): Array of per time step input delta tensors.
                Each element in the array is a single time step view into one
                tensor containing all of the time steps in sequence.
            delta_s (Array): Array of per time step input delta tensors.
                Each element in the array is a single time step view into one
                tensor containing all of the time steps in sequence.
            h_s (Tensor): Array of per time step hidden state tensors. Each
                element in the array is a single time step view into one tensor
                containing all of the time steps in sequence.
            nout (integer): Number of output units for the layer.
            num_steps (integer): Total number of time steps in the buffer.
            num_used_steps (integer): Number of time steps being used for real
                data.
            activation (Transform): Activation function for the layer.
            reverse (boolean): When true, unrolling will iterate over time steps
                in reverse (default case for RNN).
        """
        if num_used_steps is not None and num_used_steps < num_steps:
            h_s = h_s[:num_used_steps]

        if reverse:
            steps = reversed(list(zip(h_s, delta_s, delta_prev_s)))
        else:
            steps = zip(h_s, delta_s, delta_prev_s)

        for (hs, in_deltas, prev_in_deltas) in steps:
            if not math_cpu.act_and_mul(in_deltas, hs, activation):
                in_deltas[:] = activation.bprop(hs) * in_deltas
            self.compound_dot(W_recur, in_deltas, prev_in_deltas, beta=1.0)

    def allocate_new_deltas(self, delta, in_shape, parallelism):
        """
        For MKL backends, allocate new deltas for broadcast
        """
        return self.iobuf(in_shape, parallelism=parallelism)

    def allocate_new_outputs(self, layer, share_output):
        layer.allocate()

    def change_data_store_order(self, a, a_row, a_col, a_len, axis=1, b=None):
        return math_cpu.change_data_store_order(a, a_row, a_col, a_len, axis=1, b=b)

    def trans2d(self, W_recur_f, W_recur_b, W_recur_f_contiguous, W_recur_b_contiguous):
        math_cpu.trans2d(W_recur_f, W_recur_b, W_recur_f_contiguous, W_recur_b_contiguous)

    def bibnrnn_layer(self, h_buffer_all, h_ff_buffer, W_recur_f, W_recur_b, nsteps, nout):
        """
        Create a new BiBNRNN parameter object. To change the data storage type
        This then is passed as an argument to all the BiBNRNN operations.

        N: Number of images in mini-batch
        C: Number of output feature maps
        K: Number of input feature maps


        """
        return layer_mkl.BiBNRNNLayerMKL(h_buffer_all, h_ff_buffer, W_recur_f, W_recur_b, nsteps,
                                         nout)

    def compound_rnn_unroll_fprop_bibnrnn(self, ngLayer, h_buffer_all, h_ff_buffer, W_recur_f,
                                          h_prev_not_used_in_mkl, h_ff_f_not_used_in_mkl,
                                          h_f_not_used_in_mkl, b_f, W_recur_b,
                                          h_next_not_used_in_mkl, h_ff_b_not_used_in_mkl,
                                          h_b_not_used_in_mkl, b_b, nout, nsteps, nsteps_used,
                                          activation):
        self.change_data_store_order(h_buffer_all, h_buffer_all.shape[0], nsteps + 2,
                                     h_buffer_all.shape[1] // (nsteps + 2),
                                     b=ngLayer.h_all_contiguous)
        self.change_data_store_order(h_ff_buffer, h_ff_buffer.shape[0], nsteps,
                                     h_ff_buffer.shape[1] // nsteps,
                                     b=ngLayer.h_ff_buffer_contiguous)
        self.compound_rnn_unroll_fprop(W_recur_f, ngLayer.h_prev_contiguous,
                                       ngLayer.h_ff_f_contiguous, ngLayer.h_f_contiguous, b_f,
                                       nout, nsteps, nsteps_used, activation, False)

        self.compound_rnn_unroll_fprop(W_recur_b, ngLayer.h_next_contiguous,
                                       ngLayer.h_ff_b_contiguous, ngLayer.h_b_contiguous, b_b,
                                       nout, nsteps, nsteps_used, activation, True)

        self.change_data_store_order(ngLayer.h_all_contiguous, nsteps + 2,
                                     ngLayer.h_all_contiguous.shape[0],
                                     ngLayer.h_all_contiguous.shape[1] // (nsteps + 2),
                                     b=h_buffer_all)
        self.change_data_store_order(ngLayer.h_ff_buffer_contiguous, nsteps,
                                     ngLayer.h_ff_buffer_contiguous.shape[0],
                                     ngLayer.h_ff_buffer_contiguous.shape[1] // nsteps,
                                     b=h_ff_buffer)

    def compound_rnn_unroll_bprop_bibnrnn(self, ngLayer, error, in_deltas_f_not_used_in_mkl,
                                          prev_in_deltas_not_used_in_mkl,
                                          in_deltas_b_not_used_in_mkl,
                                          next_in_deltas_not_used_in_mkl, W_recur_f, W_recur_b,
                                          h_f_not_used_in_mkl,
                                          h_b_not_used_in_mkl, nout, nsteps, nsteps_used,
                                          activation, h_buffer_all):
        self.change_data_store_order(error, error.shape[0], nsteps, error.shape[1] // nsteps,
                                     b=ngLayer.error_contiguous)
        self.trans2d(W_recur_f, W_recur_b, ngLayer.W_recur_f_T_contiguous,
                     ngLayer.W_recur_b_T_contiguous)
        self.compound_rnn_unroll_bprop(ngLayer.W_recur_f_T_contiguous, ngLayer.prev_in_deltas,
                                       ngLayer.in_deltas_f, ngLayer.h_f_contiguous,
                                       nout, nsteps, nsteps_used, activation, True)

        self.compound_rnn_unroll_bprop(ngLayer.W_recur_b_T_contiguous, ngLayer.next_in_deltas,
                                       ngLayer.in_deltas_b, ngLayer.h_b_contiguous,
                                       nout, nsteps, nsteps_used, activation, False)

        self.change_data_store_order(ngLayer.h_all_contiguous, nsteps + 2,
                                     ngLayer.h_all_contiguous.shape[0],
                                     ngLayer.h_all_contiguous.shape[1] // (nsteps + 2),
                                     b=h_buffer_all)
        self.change_data_store_order(ngLayer.error_contiguous, nsteps,
                                     ngLayer.error_contiguous.shape[0],
                                     ngLayer.error_contiguous.shape[1] // nsteps, b=error)

    def detectionOutput_fprop(self, conf_view, loc_view, detection, prior_boxes,
                              proposals, nms_top_k, image_top_k, score_threshold, nms_threshold):
        conf = c_longlong(conf_view._tensor.ctypes.data)
        loc = c_longlong(loc_view._tensor.ctypes.data)
        detection = c_longlong(detection._tensor.ctypes.data)
        prior_boxes = c_longlong(prior_boxes._tensor.ctypes.data)
        L, num_class, bs = conf_view.shape
        proposals = c_longlong(proposals._tensor.ctypes.data)
        result = np.zeros((bs, image_top_k, 6), dtype=np.float32)
        result_ptr = c_longlong(result.ctypes.data)
        result_len = np.zeros(bs, dtype=np.int64)
        result_len_ptr = c_longlong(result_len.ctypes.data)

        self.mklEngine.detection_fprop(conf, loc, result_ptr, prior_boxes,
                                       result_len_ptr, c_longlong(L), c_longlong(num_class),
                                       c_longlong(bs), c_longlong(nms_top_k),
                                       c_longlong(image_top_k),
                                       c_float(score_threshold),
                                       c_float(nms_threshold))
        batch_all_detections = [None] * self.bsz
        for i in range(bs):
            leng = np.long(result_len[i])
            res_batch = np.zeros((leng, 6))
            res_batch[:] = result[i, 0:leng, :]
            batch_all_detections[i] = res_batch
        return batch_all_detections



# ----------------------------------------------------------------------------
# Copyright 2014-2016 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
"""
Contains various functions and wrappers to make code Python 2 and Python 3
compatible.
"""
from future import standard_library
standard_library.install_aliases()  # triggers E402, hence noqa below

import sys  # noqa
import logging  # noqa


logger = logging.getLogger(__name__)
PY3 = (sys.version_info[0] >= 3)

if PY3:
    xrange = range
else:
    xrange = xrange  # pylint: disable=xrange-builtin

if not PY3:
    import cPickle as the_pickle  # noqa
else:
    import pickle as the_pickle  # noqa
pickle = the_pickle


def pickle_load(filepath):
    """
    Py2Py3 compatible Pickle load

    Arguments:
        filepath (str): File containing pickle data stream to load

    Returns:
        Unpickled object
    """
    if PY3:
        return pickle.load(filepath, encoding='latin1')
    else:
        return pickle.load(filepath)



# ----------------------------------------------------------------------------
# Copyright 2017 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
from __future__ import division
import logging
import numpy as np
from PIL import Image
from neon.callbacks.callbacks import Callback
logger = logging.getLogger(__name__)

try:
    import matplotlib
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt
except ImportError as e:
    logger.exception("GANPlotCallback requires matplotlib to be installed")
    raise(e)
try:
    from scipy.signal import medfilt
except ImportError as e:
    logger.exception("GANPlotCallback requires scipy to be installed")
    raise(e)


class GANPlotCallback(Callback):
    """
    Create PNG plots of samples from a GAN model.

    Arguments:
        filename (string): Filename prefix for output PNGs.
        hw (int): Height and width of the images.
        nchan (int): number of channels.
        num_saples (int): how many samples to show from traning and generated data.
        sym_range (bool): pixel value [-1, 1] or [0, 1].
        padding (int): number of pixels to pad in output images.
        plot_width (int): width of output images.
        plot_height (int): height of output images.
        dpi (float): dots per inch.
        font_size (int): font size of labels.
        epoch_freq (int): number of epochs per plotting callback.
    """
    def __init__(self, filename, hw=28, nchan=1, num_samples=16, sym_range=False, padding=2,
                 plot_width=1200, plot_height=600, dpi=60., font_size=10, epoch_freq=1):
        super(GANPlotCallback, self).__init__(epoch_freq=epoch_freq)
        self.filename = filename
        self.hw = hw
        self.nchan = nchan
        self.num_samples = num_samples
        self.padding = padding
        self.sym_range = sym_range
        self.plot_width = plot_width
        self.plot_height = plot_height
        self.dpi = dpi
        self.font_size = font_size

    def _value_transform(self, batch):
        if self.nchan == 1:
            batch = 1. - batch
        else:
            if self.sym_range:
                batch = (batch + 1.) / 2.
        return batch

    def _shape_transform(self, batch):
        assert self.nchan * self.hw * self.hw == batch.shape[0], "wrong image size specified"
        assert self.num_samples <= batch.shape[1], "number of samples must not exceed batch size"

        nrow = int(np.ceil(np.sqrt(self.num_samples)))
        ncol = int(np.ceil(1. * self.num_samples / nrow))
        width = ncol*(self.hw+self.padding)-self.padding
        height = nrow*(self.hw+self.padding)-self.padding

        batch = batch[:, :self.num_samples]
        batch = batch.reshape(self.nchan, self.hw, self.hw, self.num_samples)
        batch = np.swapaxes(np.swapaxes(batch, 0, 2), 0, 1)

        canvas = np.ones([height, width, self.nchan])
        for i in range(self.num_samples):
            irow, icol, step = i % nrow, i // nrow, self.hw + self.padding
            canvas[irow*step:irow*step+self.hw, icol*step:icol*step+self.hw, :] = \
                batch[:, :, ::-1, i]
        if self.nchan == 1:
            canvas = canvas.reshape(height, width)
        return canvas

    def on_epoch_end(self, callback_data, model, epoch):
        # convert to numpy arrays
        data_batch = model.data_batch.get()
        noise_batch = model.noise_batch.get()
        # value transform
        data_batch = self._value_transform(data_batch)
        noise_batch = self._value_transform(noise_batch)
        # shape transform
        data_canvas = self._shape_transform(data_batch)
        noise_canvas = self._shape_transform(noise_batch)
        # plotting options
        im_args = dict(interpolation="nearest", vmin=0., vmax=1.)
        if self.nchan == 1:
            im_args['cmap'] = plt.get_cmap("gray")
        fname = self.filename+'_data_'+'{:03d}'.format(epoch)+'.png'
        Image.fromarray(np.uint8(data_canvas*255)).convert('RGB').save(fname)
        fname = self.filename+'_noise_'+'{:03d}'.format(epoch)+'.png'
        Image.fromarray(np.uint8(noise_canvas*255)).convert('RGB').save(fname)

        # plot logged WGAN costs if logged
        if model.cost.costfunc.func == 'wasserstein':
            giter = callback_data['gan/gen_iter'][:]
            nonzeros = np.where(giter)
            giter = giter[nonzeros]
            cost_dis = callback_data['gan/cost_dis'][:][nonzeros]
            w_dist = medfilt(np.array(-cost_dis, dtype='float64'), kernel_size=101)
            plt.figure(figsize=(400/self.dpi, 300/self.dpi), dpi=self.dpi)
            plt.plot(giter, -cost_dis, 'k-', lw=0.25)
            plt.plot(giter, w_dist, 'r-', lw=2.)
            plt.title(self.filename, fontsize=self.font_size)
            plt.xlabel("Generator Iterations", fontsize=self.font_size)
            plt.ylabel("Wasserstein estimate", fontsize=self.font_size)
            plt.margins(0, 0, tight=True)
            plt.savefig(self.filename+'_training.png', bbox_inches='tight')
            plt.close()



# ----------------------------------------------------------------------------
# Copyright 2014-2016 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
"""
For machine generated datasets.
"""
import numpy as np

from neon import NervanaObject


class Task(NervanaObject):
    """
    Base class from which ticker tasks inherit.
    """

    def fetch_io(self, time_steps):
        """
        Generate inputs, outputs numpy tensor pair of size appropriate for this minibatch.

        Arguments:
            time_steps (int): Number of time steps in this minibatch.

        Returns:
            tuple: (input, output) tuple of numpy arrays.

        """
        columns = time_steps * self.be.bsz
        inputs = np.zeros((self.nin, columns))
        outputs = np.zeros((self.nout, columns))
        return inputs, outputs

    def fill_buffers(self, time_steps, inputs, outputs, in_tensor, out_tensor, mask):
        """
        Prepare data for delivery to device.

        Arguments:
            time_steps (int): Number of time steps in this minibatch.
            inputs (numpy array): Inputs numpy array
            outputs (numpy array): Outputs numpy array
            in_tensor (Tensor): Device buffer holding inputs
            out_tensor (Tensor): Device buffer holding outputs
            mask (numpy array): Device buffer for the output mask
        """
        # Put inputs and outputs, which are too small, into properly shaped arrays
        columns = time_steps * self.be.bsz
        inC = np.zeros((self.nin, self.max_columns))
        outC = np.zeros((self.nout, self.max_columns))
        inC[:, :columns] = inputs
        outC[:, :columns] = outputs

        # Copy those arrays to device
        in_tensor.set(inC)
        out_tensor.set(outC)

        # Set a mask over the unused part of the buffer
        mask[:, :columns] = 1
        mask[:, columns:] = 0


class CopyTask(Task):
    """
    Copy task from the Neural Turing Machines paper:
        http://arxiv.org/abs/1410.5401.

    This version of the task is batched.
    All sequences in the same mini-batch are the same length,
    but every new minibatch has a randomly chosen minibatch length.

    When a given minibatch has length < seq_len_max, we mask the outputs
    for time steps after time_steps_max.

    The generated data is laid out in the same way as other RNN data in neon.
    """

    def __init__(self, seq_len_max, vec_size):
        """
        Set up the attributes that ticker needs to see.

        Arguments:
            seq_len_max (int): Longest allowable sequence length
            vec_size (int): Width of the bit-vector to be copied (this was 8 in paper)
        """

        self.seq_len_max = seq_len_max
        self.vec_size = vec_size
        self.nout = self.vec_size  # output has the same dimension as the underlying bit vector
        self.nin = self.vec_size + 2  # input has more dims (for the start and stop channels)
        self.time_steps_func = lambda l: 2 * l + 2
        self.time_steps_max = 2 * self.seq_len_max + 2
        self.time_steps_max = self.time_steps_func(self.seq_len_max)
        self.max_columns = self.time_steps_max * self.be.bsz

    def synthesize(self, in_tensor, out_tensor, mask):

        """
        Create a new minibatch of ticker copy task data.

        Arguments:
            in_tensor (Tensor): Device buffer holding inputs
            out_tensor (Tensor): Device buffer holding outputs
            mask (numpy array): Device buffer for the output mask
        """

        # All sequences in a minibatch are the same length for convenience
        seq_len = np.random.randint(1, self.seq_len_max + 1)
        time_steps = self.time_steps_func(seq_len)

        # Generate intermediate buffers of the right size
        inputs, outputs = super(CopyTask, self).fetch_io(time_steps)

        # Set the start bit
        inputs[-2, :self.be.bsz] = 1

        # Generate the sequence to be copied
        seq = np.random.randint(2,
                                size=(self.vec_size,
                                      seq_len * self.be.bsz))

        # Set the stop bit
        stop_loc = self.be.bsz * (seq_len + 1)
        inputs[-1, stop_loc:stop_loc + self.be.bsz] = 1

        # Place the actual sequence to copy in inputs
        inputs[:self.vec_size, self.be.bsz:stop_loc] = seq

        # Now place that same sequence in a different place in outputs
        outputs[:, self.be.bsz * (seq_len + 2):] = seq

        # Fill the device minibatch buffers
        super(CopyTask, self).fill_buffers(time_steps, inputs, outputs,
                                           in_tensor, out_tensor, mask)


class RepeatCopyTask(Task):
    """
    Repeat Copy task from the Neural Turing Machines paper:
        http://arxiv.org/abs/1410.5401.

    See Also:
        See comments on :py:class:`~neon.data.ticker.CopyTask` class for more details.
    """

    def __init__(self, seq_len_max, repeat_count_max, vec_size):
        """
        Set up the attributes that ticker needs to see.

        Arguments:
            seq_len_max (int): Longest allowable sequence length
            repeat_count_max (int): Max number of repeats
            vec_size (int): Width of the bit-vector to be copied (was 8 in paper)
        """

        self.seq_len_max = seq_len_max
        self.repeat_count_max = seq_len_max
        self.vec_size = vec_size
        self.nout = self.vec_size + 1  # we output the sequence and a stop bit in a stop channel
        self.nin = self.vec_size + 2  # input has more dims (for the start and stop channels)

        # seq is seen once as input, repeat_count times as output, with a
        # start bit, stop bit, and output stop bit
        self.time_steps_func = lambda l, r: l * (r + 1) + 3
        self.time_steps_max = self.time_steps_func(self.seq_len_max, self.repeat_count_max)
        self.max_columns = self.time_steps_max * self.be.bsz

    def synthesize(self, in_tensor, out_tensor, mask):

        """
        Create a new minibatch of ticker repeat copy task data.

        Arguments:
            in_tensor (Tensor): Device buffer holding inputs
            out_tensor (Tensor): Device buffer holding outputs
            mask (numpy array): Device buffer for the output mask
        """
        # All sequences in a minibatch are the same length for convenience
        seq_len = np.random.randint(1, self.seq_len_max + 1)
        repeat_count = np.random.randint(1, self.repeat_count_max + 1)
        time_steps = self.time_steps_func(seq_len, repeat_count)

        # Get the minibatch specific numpy buffers
        inputs, outputs = super(RepeatCopyTask, self).fetch_io(time_steps)

        # Set the start bit
        inputs[-2, :self.be.bsz] = 1

        # Generate the sequence to be copied
        seq = np.random.randint(2,
                                size=(self.vec_size,
                                      seq_len * self.be.bsz))

        # Set the repeat count
        # TODO: should we normalize repeat count?
        stop_loc = self.be.bsz * (seq_len + 1)
        inputs[-1, stop_loc:stop_loc + self.be.bsz] = repeat_count

        # Place the actual sequence to copy in inputs
        inputs[:self.vec_size, self.be.bsz:stop_loc] = seq

        # Now place that same sequence repeat_copy times in outputs
        for i in range(repeat_count):
            start = self.be.bsz * ((i + 1) * seq_len + 2)
            stop = start + seq_len * self.be.bsz
            outputs[:-1, start:stop] = seq

        # Place the output finish bit
        outputs[-1, -self.be.bsz:] = 1

        # Fill the device minibatch buffers
        super(RepeatCopyTask, self).fill_buffers(time_steps, inputs, outputs,
                                                 in_tensor, out_tensor, mask)


class PrioritySortTask(Task):
    """
    Priority Sort task from the Neural Turing Machines paper:
        http://arxiv.org/abs/1410.5401.

    See Also:
        See comments on :py:class:`~neon.data.ticker.CopyTask` class for more details.
    """

    def __init__(self, seq_len_max, vec_size):
        """
        Set up the attributes that ticker needs to see.

        Arguments:
            seq_len_max (int): Longest allowable sequence length
            vec_size (int): Width of the bit-vector to be copied (this was 8 in paper)
        """

        self.seq_len_max = seq_len_max
        self.vec_size = vec_size
        self.nout = self.vec_size  # we output the sorted sequence, with no stop bit
        self.nin = self.vec_size + 3  # extra channels for start, stop, and priority

        # seq is seen once as input with start and stop bits
        # then we output seq in sorted order
        self.time_steps_func = lambda l: 2 * l + 2
        self.time_steps_max = self.time_steps_func(self.seq_len_max)
        self.max_columns = self.time_steps_max * self.be.bsz

    def synthesize(self, in_tensor, out_tensor, mask):

        """
        Create a new minibatch of ticker priority sort task data.

        Arguments:
            in_tensor: device buffer holding inputs
            out_tensor: device buffer holding outputs
            mask: device buffer for the output mask
        """

        # All sequences in a minibatch are the same length for convenience
        seq_len = np.random.randint(1, self.seq_len_max + 1)
        time_steps = self.time_steps_func(seq_len)

        # Get the minibatch specific numpy buffers
        inputs, outputs = super(PrioritySortTask, self).fetch_io(time_steps)

        # Set the start bit
        inputs[-3, :self.be.bsz] = 1

        # Generate the sequence to be copied
        seq = np.random.randint(2,
                                size=(self.nin,
                                      seq_len * self.be.bsz)).astype(float)

        # Zero out the start, stop, and priority channels
        seq[-3:, :] = 0

        # Generate the scalar priorities and put them in seq
        priorities = np.random.uniform(-1, 1, size=(seq_len * self.be.bsz,))
        seq[-1, :] = priorities

        # Set the stop bit
        stop_loc = self.be.bsz * (seq_len + 1)
        inputs[-2, stop_loc:stop_loc + self.be.bsz] = 1

        # Place the actual sequence to copy in inputs
        inputs[:, self.be.bsz:stop_loc] = seq

        # sort the sequences
        for i in range(self.be.bsz):
            # for every sequence in the batch

            # x <- every column in the sequence
            x = seq[:, i::self.be.bsz]

            # sort that set of columns by elt in the last row (the priority)
            x = x[:, x[-1, :].argsort()]

            # put those columns back into minibatch in the right places
            seq[:, i::self.be.bsz] = x

        outputs[:, self.be.bsz * (seq_len + 2):] = seq[:self.nout, :]

        # Fill the device minibatch buffers
        super(PrioritySortTask, self).fill_buffers(time_steps, inputs, outputs,
                                                   in_tensor, out_tensor, mask)


class Ticker(NervanaObject):
    """
    This class defines methods for generating and iterating over ticker datasets.
    """

    def reset(self):
        """
        Reset has no meaning in the context of ticker data.
        """
        pass

    def __init__(self, task):
        """
        Construct a ticker dataset object.

        Arguments:
            task: An object representing the task to be trained on
                  It contains information about input and output size,
                  sequence length, etc. It also implements a synthesize function,
                  which is used to generate the next minibatch of data.
        """

        self.task = task

        # These attributes don't make much sense in the context of tickers
        # but I suspect it will be hard to get rid of them
        self.batch_index = 0
        self.nbatches = 100
        self.ndata = self.nbatches * self.be.bsz

        # Alias these because other code relies on datasets having nin and nout
        self.nout = task.nout
        self.nin = task.nin

        # Configuration elsewhere relies on the existence of this
        self.shape = (self.nin, self.task.time_steps_max)

        # Initialize the inputs, the outputs, and the mask
        self.dev_X = self.be.iobuf((self.nin, self.task.time_steps_max))
        self.dev_y = self.be.iobuf((self.nout, self.task.time_steps_max))
        self.mask = self.be.iobuf((self.nout, self.task.time_steps_max))

    def __iter__(self):
        """
        Generator that can be used to iterate over this dataset.

        Yields:
            tuple : the next minibatch of data.

        Note:
            The second element of the tuple is itself a tuple (t,m) with:
                t: the actual target as generated by the task object
                m: the output mask to account for the difference between
                    the seq_length for this minibatch and the max seq_len,
                    which is also the number of columns in X,t, and m
        """

        self.batch_index = 0

        while self.batch_index < self.nbatches:

            # The task object writes minibatch data into buffers we pass it
            self.task.synthesize(self.dev_X, self.dev_y, self.mask)

            self.batch_index += 1

            yield self.dev_X, (self.dev_y, self.mask)



#!/usr/bin/python

# Copyright 2016 Nervana Systems Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import division
import numpy as np
#from ipdb import set_trace
from struct import pack, unpack
from neon import logger as neon_logger

def ceil_div(x, y):
    return -(-x // y)

def out_dim(S, X, padding, strides):
    return ceil_div(X - S + 1 + 2 * padding, strides)

def strip_mantissa(val):
    i = unpack('I', pack('f', val))[0] & 0x7f800000
    f = unpack('f', pack('I', i))[0]
    return f

def immediate(val):
    i = unpack('I', pack('f', val))[0] & 0x7ffff000
    f = unpack('f', pack('I', i))[0]
    return f

def quantize(ary, bits, sign=1):
    maxval = float(np.max(np.absolute(ary)))
    scale  = strip_mantissa(maxval) / float(1 << (bits - sign - 1))
    ary    = np.around(ary * (1.0 / scale)).astype(np.int64)
    return ary, np.float64(scale)

######### Direct Convolution #########

def fconv_slice(q, S, X, padding, strides):
    f1 = 0
    f2 = S - 1
    x1 = q * strides - padding
    x2 = x1 + f2
    if x1 < 0:
        f1 = -x1
        x1 = 0
    if x2 >= X:
        dif = x2 - X + 1
        f2 -= dif
        x2 -= dif
    return (slice(f1, f2 + 1), slice(x1, x2 + 1), f2 - f1 + 1)

def bconv_slice(x, S, Q, padding, strides):
    qs = x - (S - padding - 1)
    firstF = None
    for s in range(S): #TODO remove loop logic here.
        q = qs + s
        if q % strides == 0:
            q //= strides
            if q >= 0 and q < Q:
                if firstF is None:
                    firstF = s
                    firstE = q
                lastF = s
                lastE = q
    if firstF is None:
        return (slice(0,0,1), slice(0,0,1), 0)
    return (slice(firstF,lastF + 1,strides), slice(firstE,lastE + 1,1), 0)

def xprop_direct(I, F, O, padding, strides, backward=False):

    if all(x == 1 for x in F.shape[1:3]):
        C = F.shape[0]
        K = F.shape[4]
        if backward:
            # CxHWN = CxK . KxHWN
            O[:] = np.dot( F.reshape((C, -1)),   I.reshape((K, -1)) ).reshape((O.shape))
        else:
            # KxHWN = CxK.T . CxHWN
            O[:] = np.dot( F.reshape((C, -1)).T, I.reshape((C, -1)) ).reshape((O.shape))
        return

    if backward:
        # C <=> K and mirror R,S
        F = np.transpose(F[:,::-1,::-1,:], (3,1,2,0)).copy()
        xconv_slice = bconv_slice
    else:
        xconv_slice = fconv_slice

    C, Y, X, N = I.shape
    C, R, S, K = F.shape
    K, P, Q, N = O.shape

    qSlice = [ xconv_slice(q, S, X, padding[0], strides[0]) for q in range(Q) ]

    for p in range(P):
        sliceR, sliceY, _ = xconv_slice(p, R, Y, padding[1], strides[1])

        for q in range(Q):
            sliceS, sliceX, _ = qSlice[q]

            slicedF = F[:,sliceR,sliceS,:].reshape((-1, K))
            slicedI = I[:,sliceY,sliceX,:].reshape((-1, N))

            O[:,p,q,:] = np.dot( slicedF.T,  slicedI )

def updat_direct(I, E, U, padding, strides):

    C, Y, X, N = I.shape
    K, P, Q, N = E.shape
    C, R, S, K = U.shape

    if all(x == 1 for x in (R, S)):
        # CxK = CxHWN . KxHWN.T
        U[:] = np.dot( I.reshape((C, -1)), E.reshape((K, -1)).T ).reshape((U.shape))
        return

    U.fill(0.0)

    qSlice = [ fconv_slice(q, S, X, padding[0], strides[0]) for q in range(Q) ]

    for p in range(P):
        sliceR, sliceY, rlen = fconv_slice(p, R, Y, padding[1], strides[1])

        for q in range(Q):
            sliceS, sliceX, slen = qSlice[q]

            slicedI = I[:,sliceY,sliceX,:].reshape((-1, N))
            slicedE = E[:,p,q,:]

            U[:,sliceR,sliceS,:] += np.dot( slicedI,  slicedE.T ).reshape((C, rlen, slen, K))

######### Winograd Convolution #########

I_2x2_5x5 = (
    np.array([
    [ 1.0,  0.00 ],
    [ 1.0,  0.75 ],
    [ 1.0, -0.75 ],
    [ 1.0,  1.50 ],
    [ 1.0, -1.50 ],
    [ 0.0,  1.00 ]]),
)

F_2x2_5x5 = (
    np.array([
    [    64.0 / 81.0,        0.0,       0.0,      0.0,      0.0 ],
    [ -128.0 / 243.0, -32.0 / 81.0, -8.0 / 27.0, -2.0 / 9.0, -1.0 / 6.0 ],
    [ -128.0 / 243.0,  32.0 / 81.0, -8.0 / 27.0,  2.0 / 9.0, -1.0 / 6.0 ],
    [   32.0 / 243.0,  16.0 / 81.0,  8.0 / 27.0,  4.0 / 9.0,  2.0 / 3.0 ],
    [   32.0 / 243.0, -16.0 / 81.0,  8.0 / 27.0, -4.0 / 9.0,  2.0 / 3.0 ],
    [          0.0,        0.0,       0.0,      0.0,      1.0 ]]),
)

O_2x2_5x5 = (
    np.array([
    [  1.265625,  0.0,     0.0,     0.0,     0.0,      0.0      ],
    [  0.0,      -1.6875,  1.6875, -0.84375, 0.84375,  1.265625 ],
    [ -2.8125,   -2.25,   -2.25,   -0.5625, -0.5625,   0.0      ],
    [  0.0,       0.75,   -0.75,    1.5,    -1.5,     -2.8125   ],
    [  1.0,       1.0,     1.0,     1.0,     1.0,      0.0      ],
    [  0.0,       0.0,     0.0,     0.0,     0.0,      1.0      ]]),
)

def trans_I_2x2_5x5(Iw, I, minimal=False, trans=False):
    if minimal:

        T0 = np.empty((6,2))
        T1 = np.empty((6,6))

        for O, I in ((T0, I), (T1, T0.T)):

            # 4*2 + 4*6 = 30

            O[0,:] = I[0,:]
            O[1,:] = I[0,:] + I[1,:] * 0.75
            O[2,:] = I[0,:] - I[1,:] * 0.75
            O[3,:] = I[0,:] + I[1,:] * 1.50
            O[4,:] = I[0,:] - I[1,:] * 1.50
            O[5,:] = I[1,:]

        Iw[:] = T1.T

    else:
        Iw[:] = np.dot( np.dot(I_2x2_5x5[trans[0]], I), I_2x2_5x5[trans[1]].T )


def trans_F_2x2_5x5(Fw, F, minimal=False, trans=False):
    if minimal:

        T0 = np.empty((6,5))
        T1 = np.empty((6,6))

        for O, I in ((T0, F[::-1,::-1]), (T1, T0.T)):

            # 14*5 + 14*6 = 154

            t0 = I[2,:] * 8.0 / 27.0
            t1 = I[1,:] * 32.0 / 81.0   + I[3,:] * 2.0 / 9.0
            t2 = I[1,:] * 16.0 / 81.0   + I[3,:] * 4.0 / 9.0
            t3 = I[0,:] * -128.0 / 243.0 - I[4,:] * 1.0 / 6.0 - t0
            t4 = I[0,:] * 32.0 / 243.0  + I[4,:] * 2.0 / 3.0 + t0

            O[0,:] = I[0,:] * 64.0 / 81.0
            O[1,:] = t3 - t1
            O[2,:] = t3 + t1
            O[3,:] = t4 + t2
            O[4,:] = t4 - t2
            O[5,:] = I[4,:]

        Fw[:] = T1.T

    else:
        Fw[:] = np.dot( np.dot(F_2x2_5x5[trans[0]], F[::-1,::-1]), F_2x2_5x5[trans[1]].T )

def trans_O_2x2_5x5(Mw, minimal=False, trans=False):
    if minimal:

        T0 = np.empty((6,6))
        T1 = np.empty((6,6))

        for O, I in ((T0, Mw), (T1, T0.T)):

            # 16*6 + 16*6 = 192

            t0 = I[1,:] + I[2,:]
            t1 = I[3,:] + I[4,:]
            t2 = I[1,:] - I[2,:]
            t3 = I[3,:] - I[4,:]

            O[0,:] = I[0,:] * 1.265625
            O[4,:] = I[0,:] + t0  + t1
            O[2,:] = t0 * -2.25   + t1 * -0.5625  + I[0,:] * -2.8125
            O[1,:] = t2 * -1.6875 + t3 * -0.84375 + I[5,:] *  1.265625
            O[3,:] = t2 *  0.75   + t3 *  1.5     + I[5,:] * -2.8125
            O[5,:] = I[5,:]

        return T1.T

    else:
        return np.dot( np.dot(O_2x2_5x5[trans[0]], Mw), O_2x2_5x5[trans[1]].T )


def image_slice(x, X, B):
    x0 = x * B
    x1 = x0 + B
    if x1 > X:
        return slice(x0,X,1), (0,1)
    return slice(x0,x1,1), (0,0)

def output_slice(x, P, B, D, pad):
    p0 = x * B + pad - 4
    p1 = p0 + D
    if p0 < 0:
        m0 = -p0
        p0 = 0
    else:
        m0 = 0
    if p1 > P:
        m1 = D - (p1 - P)
        p1 = P
    else:
        m1 = D
    return slice(p0,p1,1), slice(m0,m1,1)

def xprop_winograd(I, F, O, padding, minimal=False, trans=False, backward=False):

    if backward:
        # C <=> K and mirror R,S
        F = np.transpose(F[:,::-1,::-1,:], (3,1,2,0)).copy()
        # Invert padding
        padding = [4 - p for p in padding]

    C, Y, X, N = I.shape
    K, P, Q, N = O.shape

    B = 2
    D = 6
    Yw = ceil_div(Y, B)
    Xw = ceil_div(X, B)

    Fw = np.empty((D,D,C,K))
    Iw = np.empty((D,D,C,Yw,Xw,N))
    Mw = np.empty((D,D,K,Yw,Xw,N)) #, dtype=np.int64
    O.fill(0.0)

    # Transform Filters
    for c in range(C):
        for k in range(K):
            trans_F_2x2_5x5(Fw[:,:,c,k], F[c,:,:,k], minimal, trans)

    # Iterate over image transform dimensions and slice out tiles of the image
    for y in range(Yw):
        slice_y, pad_y = image_slice(y, Y, B)

        for x in range(Xw):
            slice_x, pad_x = image_slice(x, X, B)

            sliceI = I[:, slice_y, slice_x, :]

            # add any zero padding if needed
            if pad_y[1] or pad_x[1]:
                sliceI = np.pad(sliceI, ((0,0), pad_y, pad_x, (0,0)), 'constant')

            # Apply the Image transform
            for c in range(C):
                for n in range(N):
                    trans_I_2x2_5x5(Iw[:,:,c,y,x,n], sliceI[c,:,:,n], minimal, trans)

    #print Iw[:,:,0,0,0,0]
    # print Fw[:,:,0,0].reshape(36,1)
    #exit()

    # Fw, scaleF = quantize(Fw, 16)
    # Iw, scaleI = quantize(Iw, 16)

    # Fw = Fw.astype(np.float16).astype(np.float64)
    # Iw = Iw.astype(np.float32).astype(np.float64)

    # Batched gemm for the pointwise multiplication step
    for s in range(D):
        for t in range(D):
            # [K,Yw,Xw,N] = [C,K].T . [C,YwXwN]
            Mw[s,t] = np.dot( Fw[s,t].T, Iw[s,t].reshape(C, -1) ).reshape((K,Yw,Xw,N))

    #print Mw[:,:,0,0,0,0]
    #exit()

    # Mw = Mw.astype(np.float64) * scaleF * scaleI

    # Iterate over the convovled result in the pointwise space and apply inverse transform
    for y in range(Yw):
        slice_p, slice_y = output_slice(y, P, B, D, padding[0])
        for x in range(Xw):
            slice_q, slice_x = output_slice(x, Q, B, D, padding[1])
            for k in range(K):
                for n in range(N):
                    # Toss out any points that don't fit
                    Out = trans_O_2x2_5x5(Mw[:,:,k,y,x,n], minimal, trans)

                    O[k,slice_p,slice_q,n] += Out[slice_y,slice_x]


### Test Code ###

np.set_printoptions(threshold=8192 * 4, linewidth=600, formatter={'float':lambda x: "%4.0f" % x})

minimal = 1
trans = (0,0)
ones = 0
N    = 4
C, K = 4, 4
Y, X = 6, 6
R, S = 5, 5     # Fixed winograd dim
strides = 1, 1  # Fixed winograd dim
padding = 2, 2  # 0-2

P = out_dim(R, Y, padding[0], strides[0])
Q = out_dim(S, X, padding[1], strides[1])

neon_logger.display("{}".format(P, Q))

dimI = (C,Y,X,N)
dimF = (C,R,S,K)
dimO = (K,P,Q,N)

if ones:
    I  = np.ones(dimI)
    F  = np.ones(dimF)
    E  = np.ones(dimO)

    # for c in range(2):
    #     for n in range(32):
    #         I[c,:,:,n] = np.arange(0,36, dtype=np.float32).reshape(6,6)

    #I[0,:,:,0] = np.arange(0,36).reshape(6,6)
    F[0,:,:,0] = np.arange(0,25).reshape(5,5)

    # for p,q in np.ndindex((Y,X)):
    #     I[:,p,q,:] = np.identity(N)

    # for p,q in np.ndindex((P,Q)):
    #     for n in range(N):
    #         E[:,p,q,n] = range(K)

    # for c in range(C):
    #     for n in range(N):
    #         I[c,:,:,n] = c+1 #np.arange(1+c,37+c).reshape((6,6))

    # for k in range(K):
    #     for n in range(N):
    #         E[k,:,:,n] = k+1 #np.arange(1+k,17+k).reshape((4,4))

else:
    I  = np.random.uniform(-1.0, 1.0, dimI)
    #F  = np.random.normal(0.0, 0.1, dimF)
    F  = np.random.uniform(-1.0, 1.0, dimF)
    E  = np.random.uniform(-1.0, 1.0, dimO)

Od = np.empty(dimO)
Ow = np.empty(dimO) #, dtype=np.float32

Bd = np.empty(dimI)
Bw = np.empty(dimI) #, dtype=np.float32

xprop_direct(I, F, Od, padding, strides)
xprop_winograd(I, F, Ow, padding, minimal=minimal, trans=trans)

xprop_direct(E, F, Bd, padding, strides, backward=True)
xprop_winograd(E, F, Bw, padding, minimal=minimal, trans=trans, backward=True)

difO = Od - Ow
difB = Bd - Bw

neon_logger.display(abs(difO).max() / Od.max())
neon_logger.display(abs(difB).max() / Bd.max())

# print Od[0,:,:,0]
# print Ow[0,:,:,0]
# print difO[0,:,:,0]



# ----------------------------------------------------------------------------
# Copyright 2014-2016 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
"""
Nervana's deep learning library
"""
from __future__ import print_function
from builtins import zip

try:
    from neon.version import VERSION as __version__  # noqa
except ImportError:
    import sys
    print("ERROR: Version information not found.  Ensure you have built "
          "the software.\n    From the top level dir issue: 'make'")
    sys.exit(1)
from copy import deepcopy
import inspect
import logging

from neon.util.persist import load_class


DISPLAY_LEVEL_NUM = 41
logging.addLevelName(DISPLAY_LEVEL_NUM, "DISPLAY")


def display(self, message, *args, **kwargs):
    if self.isEnabledFor(DISPLAY_LEVEL_NUM):
        self._log(DISPLAY_LEVEL_NUM, message, args, **kwargs)

logging.Logger.display = display

# setup a preliminary stream based logger
logging.basicConfig(level=logging.ERROR)
logger = logging.getLogger(__name__)


def get_args(func):
    """
    returns a dictionary of arg_name:default_values for the input function
    """
    args, varargs, keywords, defaults = inspect.getargspec(func)

    defaults = list(reversed(defaults)) if defaults is not None else []
    args = list(reversed(args))

    while len(defaults) != len(args):
        defaults += (None,)

    return dict(list(zip(args, defaults)))


class NervanaObject(object):
    """
    Base (global) object available to all other classes.

    Attributes:
        be (Backend): Hardware backend being used.
    """
    be = None
    __counter = 0

    def __init__(self, name=None):
        """
        Class constructor.

        Args:
            name (str, optional): Name to assign instance of this class.
        """
        if name is None:
            name = '{}_{}'.format(self.classnm, self.__counter)
        self.name = name
        self._desc = None
        type(self).__counter += 1

    @staticmethod
    def recursive_gen(pdict, key):
        """
        helper method to check whether the definition
        dictionary is defining a NervanaObject child,
        if so it will instantiate that object and replace
        the dictionary element with an instance of that object
        """
        if type(pdict[key]) is dict and 'type' in pdict[key]:
            if 'config' not in pdict[key]:
                # assume empty params if 'config' key missing
                # needed to keep yaml deserialization short
                pdict[key]['config'] = {}
            ccls = load_class(pdict[key]['type'])
            pdict[key] = ccls.gen_class(pdict[key]['config'])

    @classmethod
    def gen_class(cls, pdict):
        for key in pdict:
            # check for any nervanaobject types in
            # the parameter list
            if type(pdict[key]) in [tuple, list]:
                for ind, val in enumerate(pdict[key]):
                    cls.recursive_gen(pdict[key], ind) 
            else:
                cls.recursive_gen(pdict, key)

        return cls(**pdict)

    def __del__(self):
        type(self).__counter -= 1

    @property
    def classnm(self):
        """
        Returns the class name.
        """
        return self.__class__.__name__

    @property
    def modulenm(self):
        """
        Returns the full module path.
        """
        return self.__class__.__module__ + '.' + self.__class__.__name__

    def get_description(self, skip=[], **kwargs):
        """
        Returns a ``dict`` that contains all necessary information needed
        to serialize this object.

        Arguments:
            skip (list): Objects to omit from the dictionary.

        Returns:
            (dict): Dictionary format for object information.
        """
        if type(skip) is not list:
            skip = list(skip)
        else:
            skip = deepcopy(skip)
        skip.append('self')

        config = {}
        defaults = get_args(self.__init__)
        for arg in defaults:
            if arg in skip:
                continue

            # all args need to go in the __dict__ so we can read
            # them out the way they were read in. alternatively,
            # you can override get_description to say how to
            # put them in to the description dictionary.
            if arg in self.__dict__:
                if self.__dict__[arg] != defaults[arg]:
                    if isinstance(self.__dict__[arg], NervanaObject):
                        config[arg] = self.__dict__[arg].get_description()
                    else:
                        config[arg] = self.__dict__[arg]
            else:
                logger.warning("can't describe argument '{}' to {}".format(arg, self))

        desc = {'type': self.modulenm, 'config': config}
        self._desc = desc
        return desc



# ----------------------------------------------------------------------------
# Copyright 2014 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
"""
Defines basic input datatset types.
"""
import os
import h5py
import logging
import numpy as np

from neon.data import ArrayIterator
logger = logging.getLogger(__name__)


class HDF5Iterator(ArrayIterator):
    """
    Data iterator which uses an HDF5 file as the source of the data, useful when
    the entire HDF5 dataset cannot fit into memory (for smaller datasets, use the ArrayIterator).

    To initialize the HDF5Iterator, simply call::

        train_set = HDF5Iterator('your_data_path.h5')

    The HDF5 file format must contain the following datasets:

    - `input` (ndarray):
                        Input data, which is a 2-D array (float or uint8) that
                        has size `(N, F)`, where `N` is the number of examples,
                        and `F` is the number of features. For images, `F = C*H*W` where
                        `C` is the number of channels and `H` and `W` are the height
                        and width of the image, respectively. This data must also
                        have the following attributes:
        - `lshape` (tuple):
                    Tuple of ints indicating the shape of each
                    input (for examples, image data may have
                    an lshape of `[C, H, W]`)
        - `mean` (ndarray):
                    The mean to subtract, either formatted as (F, 1) or
                    a mean for each channel with dimensions (C, 1)
    - `output` (ndarray):
                        An optional dataset which, if supplied, will be
                        used at the target/expected output of the network. the
                        array should have the shape `(N, M)` where `N` is the number
                        of items (must match the `N` dim of the input set)
                        and `M` is the size of the output data which must match
                        size of ouput from the output layer of the network.

    For cases where the output should be converted to a one-hot encoding (see Loading Data),
    use the `HDF5IteratorOneHot`. Or for autoencoder problems, use `HDFIteratorAutoencoder`.
    """
    def __init__(self, hdf_filename, name=None):
        """
        Args:
            hdf_filename (string): Path to the HDF5 datafile.
            name (string, optional): Name to assign this iterator. Defaults to None.
        """
        super(ArrayIterator, self).__init__(name=name)

        self.hdf_filename = hdf_filename

        if not os.path.isfile(hdf_filename):
            raise IOError('File not found %s' % hdf_filename)
        self.hdf_file = h5py.File(hdf_filename, mode='r', driver=None)

        # input data array
        self.inp = self.hdf_file['input']
        self.ndata = self.inp.shape[0]

        # must have at least 1 minibatch of data in the file
        assert self.ndata >= self.be.bsz
        self.start = 0

        # the input array unflattened size
        self.lshape = tuple(self.inp.attrs['lshape'])
        self.shape = self.lshape

        if 'output' in self.hdf_file:
            self.out = self.hdf_file['output']

        self.inpbuf = None
        self.outbuf = None
        self.allocated = False

    def allocate(self):
        """
        After the input and output (`self.inp` and `self.out)` have been
        set this function will allocate the host and device buffers
        for the mini-batches.

        The host buffer is referenced as `self.mini_batch_in` and `self.mini_batch_out`, and
        stored on device as `self.inbuf` and `self.outbuf`.
        """
        if not self.allocated:
            self.allocate_inputs()
            self.allocate_outputs()
            self.allocated = True

    def allocate_inputs(self):
        """
        Allocates the host and device input data buffers
        and any other associated storage.

        `self.inpbuf` is the on-device buffer for the input minibatch
        `self.mini_batch_in` is the on-host buffer for the input minibatch
        `self.mean` is the on-device buffer of the mean array
        """
        # on device minibatch_buffer (input)
        self.inpbuf = self.be.iobuf(self.inp.shape[1])

        # setup host buffer for a mini_batch
        self.mini_batch_in = np.zeros(self.inpbuf.shape)

        self.mean = None
        # the 'mean' dataset is the the mean values to subtract
        if 'mean' in self.hdf_file:
            mns_ = np.array(self.hdf_file['mean']).flatten()
            if mns_.size != self.inp.shape[1]:
                # channel by channel mean
                # there should be 1 element per channel
                assert mns_.size == self.lshape[0], 'mean image size mismatch'
                # need to have 2-d array for broadcasting
                mns_ = mns_.reshape((self.lshape[0], 1)).copy()
                # make channel-by-channel mean subtraction view
                self.meansub_view = self.inpbuf.reshape((self.lshape[0], -1))
            else:
                self.meansub_view = self.inpbuf

            self.mean = self.be.array(mns_)

    def allocate_outputs(self):
        """
        Allocates the host and device output data buffers
        and any other associated storage.

        `self.outbuf` is the on-device buffer for the output minibatch
        `self.mini_batch_out` is the on-host buffer for the output minibatch
        """
        self.outbuf = None
        if 'output' in self.hdf_file:
            self.outbuf = self.be.iobuf(self.out.shape[1])
            self.mini_batch_out = np.zeros(self.outbuf.shape)

    def gen_input(self, mini_batch):
        """
        Function to handle any preprocessing before pushing an input
        mini-batch to the device.  For example, mean subtraction etc.

        Arguments:
            mini_batch (ndarray): M-by-N array where M is the flatten
                                  input vector size and N is the batch size
        """
        self.inpbuf[:] = mini_batch
        # mean subtract
        if self.mean is not None:
            self.meansub_view[:] = -self.mean + self.meansub_view

    def gen_output(self, mini_batch):
        """
        Function to handle any preprocessing before pushing an output
        mini-batch to the device.  For example, one-hot generation.

        Arguments:
            mini_batch (ndarray): M-by-N array where M is the flatten
                                  output vector size and N is the batch size
        """
        self.outbuf[:] = mini_batch

    def __del__(self):
        self.cleanup()
        # needed for python3 to exit cleanly
        super(HDF5Iterator, self).__del__()

    def cleanup(self):
        """
        Closes the HDF file.
        """
        self.hdf_file.close()

    def reset(self):
        """
        Resets the index to zero.
        """
        self.start = 0

    def __iter__(self):
        """
        Defines a generator that can be used to iterate over this dataset.

        Yields:
            tuple: The next minibatch. A minibatch includes both features and
            labels.
        """
        if not self.allocated:
            self.allocate()
        full_shape = list(self.lshape)
        full_shape.append(-1)

        mini_batch_in = self.mini_batch_in
        if self.outbuf is not None:
            mini_batch_out = self.mini_batch_out
        for i1 in range(self.start, self.ndata, self.be.bsz):
            i2 = min(i1 + self.be.bsz, self.ndata)
            bsz = i2 - i1
            if i2 == self.ndata:
                self.start = self.be.bsz - bsz

            # load mini batch on host
            xdev = self.inp
            mini_batch_in[:, :bsz] = xdev[i1:i2, :].T.astype(np.float32)
            if self.be.bsz > bsz:
                mini_batch_in[:, bsz:] = xdev[:(self.be.bsz - bsz), :].T.astype(np.float32)

            # push to device
            self.gen_input(mini_batch_in)

            if self.outbuf is not None:
                mini_batch_out[:, :bsz] = self.out[i1:i2].T
                if self.be.bsz > bsz:
                    mini_batch_out[:, bsz:] = self.out[:(self.be.bsz - bsz)].T

                self.gen_output(mini_batch_out)

            inputs = self.inpbuf
            targets = self.outbuf
            yield (inputs, targets)


class HDF5IteratorOneHot(HDF5Iterator):
    """
    Extends the HDF5Iterator class to add one hot conversion of the
    target data. For example::

        train_set = HDF5IteratorOneHot('your_data_path.h5')

    The HDF5 file format must contain the following datasets:
    - `input` dataset: Input data, which is a 2-D array (float or uint8) that
                        has size `(N, C*H*W)`, where `N` is the number of examples,
                        `C` is the number of channels, and `H` and `W` are the height
                        and width of the image, respectively. This dataset must also
                        have the following attributes:
        - `lshape`: a tuple of int indicating the shape of each
                    input (for examples, image data may have
                    an lshape of `[C, H, W]`)
        - `mean`: the mean to subtract, either formatted as (C*H*W, 1) or
                  a mean for each channel with dimensions (C, 1)
    - `output` dataset: An optional dataset which, if supplied, will be
                        used at the target/expected output of the network. the
                        array should have the shape `(N, M)` where `N` is the number
                        of items (must match the `N` dim of the input set)
                        and `M` is the size of the output data which must match
                        size of ouput from the output layer of the network.

    The "output" dataset in the HDF5 (if present) must have the 'nclass'
    attribute specifying the number of total output classes which is needed
    for generating the one-hot encoding.
    """
    def __init__(self, hdf_filename, name=None):
        """
        Args:
            hdf_filename (string): Path to the HDF5 datafile.
            name (string, optional): Name to assign this iterator. Defaults to None.
        """
        super(HDF5IteratorOneHot, self).__init__(hdf_filename, name=name)
        if 'output' in self.hdf_file:
            assert 'nclass' in self.hdf_file['output'].attrs, 'Missing nclass attribute'
            self.nclass = int(self.hdf_file['output'].attrs['nclass'])
            self.out = np.array(self.hdf_file['output'], dtype=np.int32).reshape((-1, 1))

    def allocate_outputs(self):
        self.outbuf = None
        if 'output' in self.hdf_file:
            self.argmax_buf = self.be.iobuf(1, dtype=np.int32)
            self.mini_batch_out = np.zeros(self.argmax_buf.shape, dtype=np.int32)
            self.outbuf = self.be.iobuf(self.nclass)

    def gen_output(self, mini_batch):
        self.argmax_buf[:] = mini_batch
        self.be.onehot(self.argmax_buf, axis=0, out=self.outbuf)


class HDF5IteratorAutoencoder(HDF5Iterator):
    """
    Extends the base HDF5Iterator class for an autoencoder model.
    Returns the input data as the target.
    """
    def __iter__(self):
        """
        Defines a generator that can be used to iterate over this dataset.

        Yields:
            tuple: The next minibatch containing the inputs and the targets (here target=inputs)
        """
        for x, t in super(HDF5IteratorAutoencoder, self).__iter__():
            yield (x, x)



from __future__ import division
import numpy as np
from neon.data.dataloader_transformers import DataLoaderTransformer
from collections import OrderedDict


class DataLoaderAdapter(DataLoaderTransformer):
    """
    DataLoaderAdapter converts Aeon data buffers to tensors.

    Arguments:
        dataloader (DataLoader): Aeon dataloading module.
    """
    def __init__(self, dataloader):
        super(DataLoaderAdapter, self).__init__(dataloader, None)

        self.shape = self.shapes()[0]
        self.nbatches, modal = divmod(self.dataloader.ndata, self.be.bsz)
        if modal > 0:
            self.nbatches += 1
        self.outputs = OrderedDict()  # Create output buffers

        def max_dur(val, freq):
            uval = float(val.split(" ")[0])
            ucat = val.split(" ")[1]
            if(ucat == "samples"):
                return float(uval)
            elif(ucat == "seconds"):
                return float(uval * freq)
            elif(ucat == "milliseconds"):
                return float(uval/1000 * freq)
            else:
                raise ValueError("Unknown time unit " + ucat)

        # Convert max duration value for audio from Aeon format
        for conf in self.dataloader.config['etl']:
            if(conf['type'] == "audio"):
                self.max_duration = max_dur(conf['max_duration'], conf['sample_freq_hz'])

    def transform(self, t):
        """
        Converts Aeon data to tuple of tensors.

        Arguments:
            t (tuple): Tuple of numpy arrays.
                For example: {tuple}{'image':ndarray(...), 'label':ndarray(...)}
                where 'image' shape is (N,C,H,W) and 'label' shape is (N,1)
        """
        for key, value in t:
            assert value.shape[0] == self.be.bsz

            reshape_rows = self.be.bsz

            # Convert audio length from absolute to percentage
            if key == 'audio_length':
                for x in np.nditer(value, op_flags=['readwrite']):
                    x[...] = x/self.max_duration*100
                value = value.astype(np.uint8, copy=False)

            if key == 'char_map':
                x = value
                x = x[x != 0]
                x = np.reshape(x, (1, -1))
                x = np.ascontiguousarray(x)  # Contigious array needed for GPU backend
            else:
                # Adjust Aeon data layout to Neon layout, e.g. shape (N,C,H,W) -> (N,C*H*W)
                x = np.reshape(value, (reshape_rows, -1))
                x = np.ascontiguousarray(x.T)  # Contigious array needed for GPU backend

            # Patch that forces the numpy array into C order to fix the strides for NumPy > 1.11.1
            x = np.array(x, order='C')

            # Create tensor
            self.outputs[key] = self.be.array(x, dtype=value.dtype)

        return tuple(self.outputs.values())

    def shapes(self):
        # Get tuple of shape values only
        shapes = []
        for name, value in self.dataloader.axes_info:
            vals = ()
            for child_name, child_value in value:
                vals = vals + (child_value, )
            shapes.append(vals)

        return tuple(shapes)



# ----------------------------------------------------------------------------
# Copyright 2014-2016 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
"""
Functions used to load commonly available datasets.
"""
from __future__ import division
from future import standard_library
standard_library.install_aliases()  # triggers E402, hence noqa below
from future.moves.urllib.request import Request, urlopen  # noqa

import logging  # noqa
import os  # noqa
import sys  # noqa
import zipfile  # noqa

from neon import NervanaObject, logger as neon_logger  # noqa
from neon.util.compat import PY3  # noqa

logger = logging.getLogger(__name__)


class Dataset(NervanaObject):
    """
    Container class for stock datasets.

    Arguments:
        filename (str/list): name of the file to download
        url (str): url for dataset
        size (int/list): file size
        path (str): local path to place files
        subset_pct (float/int): percentage of data set to use for training
    """
    def __init__(self, filename, url, size, path='.', subset_pct=100):
        # parameters to use in dataset config serialization
        super(Dataset, self).__init__(name=None)
        self.filename = filename
        self.url = url
        self.size = size
        self.path = path
        self.subset_pct = subset_pct
        self._data_dict = None
        if subset_pct != 100:
            # placeholder to use partial data set
            raise NotImplemented('subset percentage feature is not yet implemented')

    def serialize(self):
        """
        Generates dictionary with the required parameters to describe this object
        """
        return self.get_description()

    def load_zip(self, filename, size):
        """
        Helper function for downloading test files
        Will download and unzip the file into the directory self.path

        Arguments:
            filename (str): name of file to download from self.url
            size (str): size of the file in bytes?

        Returns:
            str: Path to the downloaded dataset.
        """
        workdir, filepath = self._valid_path_append(self.path, '', filename)

        if not os.path.exists(filepath):
            self.fetch_dataset(self.url, filename, filepath, size)
        if '.zip' in filepath:
            zip_ref = zipfile.ZipFile(filepath)
            zip_ref.extractall(workdir)
            zip_ref.close()
            filepath = filepath.split('.zip')[0]
        return filepath

    @staticmethod
    def _valid_path_append(path, *args):
        """
        Helper to validate passed path directory and append any subsequent
        filename arguments.

        Arguments:
            path (str): Initial filesystem path.  Should expand to a valid
                        directory.
            *args (list, optional): Any filename or path suffices to append to path
                                    for returning.

            Returns:
                (list, str): path prepended list of files from args, or path alone if
                         no args specified.

        Raises:
            ValueError: if path is not a valid directory on this filesystem.
        """
        full_path = os.path.expanduser(path)
        res = []
        if not os.path.exists(full_path):
            os.makedirs(full_path)
        if not os.path.isdir(full_path):
            raise ValueError("path: {0} is not a valid directory".format(path))
        for suffix_path in args:
            res.append(os.path.join(full_path, suffix_path))
        if len(res) == 0:
            return path
        elif len(res) == 1:
            return res[0]
        else:
            return res

    @staticmethod
    def fetch_dataset(url, sourcefile, destfile, totalsz):
        """
        Download the file specified by the given URL.

        Args:
            url (str): Base URL of the file to be downloaded.
            sourcefile (str): Name of the source file.
            destfile (str): Path to the destination.
            totalsz (int): Size of the file to be downloaded.
        """
        req = Request(os.path.join(url, sourcefile), headers={'User-Agent': 'neon'})
        # backport https limitation and workaround per http://python-future.org/imports.html
        cloudfile = urlopen(req)
        neon_logger.display("Downloading file: {}".format(destfile))
        blockchar = u'\u2588'  # character to display in progress bar
        with open(destfile, 'wb') as f:
            data_read = 0
            chunksz = 1024**2
            while 1:
                data = cloudfile.read(chunksz)
                if not data:
                    break
                data_read = min(totalsz, data_read + chunksz)
                progress_string = u'Download Progress |{:<50}| '.format(
                    blockchar * int(float(data_read) / totalsz * 50))
                sys.stdout.write('\r')
                if PY3:
                    sys.stdout.write(progress_string)
                else:
                    sys.stdout.write(progress_string.encode("utf-8"))
                sys.stdout.flush()

                f.write(data)
            neon_logger.display("Download Complete")

    def gen_iterators(self):
        """
        Method that generates the data set iterators for the
        train, test and validation data sets.  This method
        needs to set the instance data_set attribute to a
        dictionary of data iterators.

        Returns:
            dict:  dictionary with the various data set iterators
        """
        raise NotImplemented()

    @property
    def data_dict(self):
        if self._data_dict is None:
            self._data_dict = self.gen_iterators()
        return self._data_dict

    def get_iterator(self, setname):
        """
        Helper method to get the data iterator for specified dataset

        Arguments:
            setname (str): which iterator to return (e.g. 'train', 'valid')
        """
        assert setname in self.data_dict, 'no iterator for set %s' % setname
        return self.data_dict[setname]

    @property
    def train_iter(self):
        """
        Helper method to return training set iterator
        """
        return self.get_iterator('train')

    @property
    def valid_iter(self):
        """
        Helper method to return validation set iterator
        """
        return self.get_iterator('valid')

    @property
    def test_iter(self):
        """
        Helper method to return test set iterator
        """
        return self.get_iterator('test')
